# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import webtest
from testing_support.fixtures import (
    cat_enabled,
    dt_enabled,
    override_application_settings,
    reset_core_stats_engine,
    validate_attributes,
)
from testing_support.validators.validate_browser_attributes import (
    validate_browser_attributes,
)
from testing_support.validators.validate_error_event_attributes import (
    validate_error_event_attributes,
)
from testing_support.validators.validate_error_event_attributes_outside_transaction import (
    validate_error_event_attributes_outside_transaction,
)
from testing_support.validators.validate_error_trace_attributes import (
    validate_error_trace_attributes,
)
from testing_support.validators.validate_error_trace_attributes_outside_transaction import (
    validate_error_trace_attributes_outside_transaction,
)
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_error_trace_attributes import (
    validate_transaction_error_trace_attributes,
)
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)
from testing_support.validators.validate_transaction_trace_attributes import (
    validate_transaction_trace_attributes,
)

from newrelic.api.application import application_instance as application
from newrelic.api.background_task import background_task
from newrelic.api.message_transaction import message_transaction
from newrelic.api.time_trace import notice_error
from newrelic.api.transaction import add_custom_attribute, set_user_id
from newrelic.api.wsgi_application import wsgi_application
from newrelic.common.object_names import callable_name

try:
    from testing_support.asgi_testing import AsgiTest
    from testing_support.sample_asgi_applications import normal_asgi_application
except SyntaxError:
    normal_asgi_application = None


URL_PARAM = "some_key"
URL_PARAM2 = "second_key"
REQUEST_URL = "/?" + URL_PARAM + "=someval&" + URL_PARAM2 + "=anotherval"
REQUEST_HEADERS = [
    ("Accept", "*/*"),
    ("Host", "foobar"),
    ("User-Agent", "test_attributes_in_action"),
    ("Content-Type", "text/html; charset=utf-8"),
    ("Content-Length", "10"),
]

REQ_PARAMS = ["request.parameters." + URL_PARAM, "request.parameters." + URL_PARAM2]
DISTRIBUTED_TRACE_ATTRS = [
    "traceId",
    "priority",
    "parent.type",
    "parent.app",
    "parent.account",
    "parent.transportType",
    "parent.transportDuration",
    "parentId",
    "guid",
    "sampled",
    "parentSpanId",
]

USER_ATTRS = ["puppies", "sunshine"]

TRACE_ERROR_AGENT_KEYS = [
    "response.status",
    "request.method",
    "request.headers.contentType",
    "request.uri",
    "request.headers.accept",
    "request.headers.contentLength",
    "request.headers.host",
    "request.headers.userAgent",
    "response.headers.contentLength",
    "response.headers.contentType",
]

AGENT_KEYS_ALL = TRACE_ERROR_AGENT_KEYS + REQ_PARAMS

TRANS_EVENT_INTRINSICS = (
    "name",
    "duration",
    "type",
    "timestamp",
    "totalTime",
    "error",
    "nr.apdexPerfZone",
    "apdexPerfZone",
)
TRANS_EVENT_AGENT_KEYS = [
    "response.status",
    "request.method",
    "request.uri",
    "request.headers.contentType",
    "request.headers.contentLength",
    "response.headers.contentLength",
    "response.headers.contentType",
]

BROWSER_INTRINSIC_KEYS = [
    "beacon",
    "errorBeacon",
    "licenseKey",
    "applicationID",
    "transactionName",
    "queueTime",
    "applicationTime",
    "agent",
]
ABSENT_BROWSER_KEYS = ["request.method", "request.headers.contentType", "request.headers.contentLength"]

ERROR_EVENT_INTRINSICS = (
    "type",
    "error.class",
    "error.message",
    "error.expected",
    "timestamp",
    "transactionName",
    "duration",
)
ERROR_PARAMS = [
    "ohnoes",
]
ERROR_USER_ATTRS = USER_ATTRS + ERROR_PARAMS


@wsgi_application()
def normal_wsgi_application(environ, start_response):
    status = "200 OK"

    output = "<html><head>header</head><body><p>RESPONSE</p></body></html>"
    output = output.encode("UTF-8")

    add_custom_attribute(USER_ATTRS[0], "test_value")
    add_custom_attribute(USER_ATTRS[1], "test_value")

    response_headers = [("Content-Type", "text/html; charset=utf-8"), ("Content-Length", str(len(output)))]
    start_response(status, response_headers)

    try:
        raise ValueError("Transaction had bad value")
    except ValueError:
        notice_error(attributes={ERROR_PARAMS[0]: "param-value"})

    return [output]


application_params = [normal_wsgi_application]
if normal_asgi_application:
    application_params.append(normal_asgi_application)


@pytest.fixture(scope="module", params=application_params)
def normal_application(request):
    if request.param is normal_wsgi_application:
        return webtest.TestApp(normal_wsgi_application)
    return AsgiTest(normal_asgi_application)


# Tests for checking the presence and format of agent attributes.
# Test default settings.

_expected_attributes = {"agent": TRACE_ERROR_AGENT_KEYS, "user": ERROR_USER_ATTRS, "intrinsic": ["trip_id"]}

_expected_attributes_event = {
    "agent": TRACE_ERROR_AGENT_KEYS,
    "user": ERROR_USER_ATTRS,
    "intrinsic": ERROR_EVENT_INTRINSICS,
}

_expected_absent_attributes = {"agent": REQ_PARAMS, "user": [], "intrinsic": DISTRIBUTED_TRACE_ATTRS}


@cat_enabled
@validate_error_event_attributes(_expected_attributes_event, _expected_absent_attributes)
@validate_transaction_error_trace_attributes(_expected_attributes, _expected_absent_attributes)
def test_error_in_transaction_default_settings(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_expected_attributes = {"agent": TRACE_ERROR_AGENT_KEYS, "user": USER_ATTRS, "intrinsic": ["trip_id"]}


@cat_enabled
@validate_transaction_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings({})
def test_transaction_trace_default_attribute_settings(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_expected_attributes = {"agent": TRANS_EVENT_AGENT_KEYS, "user": USER_ATTRS, "intrinsic": TRANS_EVENT_INTRINSICS}

_expected_absent_attributes = {"agent": REQ_PARAMS, "user": [], "intrinsic": []}


@validate_transaction_event_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings({})
def test_transaction_event_default_attribute_settings(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


@dt_enabled
@validate_span_events(expected_users=_expected_attributes["user"])
def test_root_span_default_attribute_settings(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Browser monitoring off by default, turn on and check default destinations


_override_settings = {"browser_monitoring.attributes.enabled": True}

_expected_attributes = {"agent": [], "user": USER_ATTRS, "intrinsic": BROWSER_INTRINSIC_KEYS}

_expected_absent_attributes = {"agent": ABSENT_BROWSER_KEYS + REQ_PARAMS}


@validate_browser_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_browser_default_attribute_settings(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test exclude request params.

_override_settings = {"error_collector.attributes.exclude": ["request.parameters.*"]}

_expected_attributes = {"agent": TRACE_ERROR_AGENT_KEYS, "user": ERROR_USER_ATTRS, "intrinsic": ["trip_id"]}

_expected_attributes_event = {
    "agent": TRACE_ERROR_AGENT_KEYS,
    "user": ERROR_USER_ATTRS,
    "intrinsic": ERROR_EVENT_INTRINSICS,
}

_expected_absent_attributes = {"agent": REQ_PARAMS, "user": [], "intrinsic": []}


@validate_error_event_attributes(_expected_attributes_event, _expected_absent_attributes)
@validate_transaction_error_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_error_in_transaction_exclude_request_params(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"transaction_tracer.attributes.exclude": ["request.parameters.*"]}

_expected_attributes = {"agent": TRACE_ERROR_AGENT_KEYS, "user": USER_ATTRS, "intrinsic": ["trip_id"]}


@validate_transaction_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_trace_exclude_request_params(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"capture_params": True, "error_collector.attributes.exclude": ["request.parameters.*"]}

_expected_attributes = {"agent": TRACE_ERROR_AGENT_KEYS, "user": USER_ATTRS, "intrinsic": ["trip_id"]}

_expected_absent_attributes = {"agent": REQ_PARAMS, "user": [], "intrinsic": []}


@validate_transaction_error_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_error_in_transaction_capture_params_exclude_request_params(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"capture_params": True, "transaction_tracer.attributes.exclude": ["request.parameters.*"]}


@validate_transaction_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_trace_capture_params_exclude_request_params(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test include request params.

_override_settings = {"error_collector.attributes.include": ["request.parameters.*"]}

_expected_attributes = {"agent": AGENT_KEYS_ALL, "user": ERROR_USER_ATTRS, "intrinsic": ["trip_id"]}

_expected_attributes_event = {"agent": AGENT_KEYS_ALL, "user": ERROR_USER_ATTRS, "intrinsic": ERROR_EVENT_INTRINSICS}


@validate_error_event_attributes(_expected_attributes_event)
@validate_transaction_error_trace_attributes(_expected_attributes)
@override_application_settings(_override_settings)
def test_error_in_transaction_include_request_params(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"transaction_tracer.attributes.include": ["request.parameters.*"]}

_expected_attributes = {"agent": AGENT_KEYS_ALL, "user": USER_ATTRS, "intrinsic": ["trip_id"]}


@validate_transaction_trace_attributes(_expected_attributes)
@override_application_settings(_override_settings)
def test_transaction_trace_include_request_params(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"transaction_events.attributes.include": ["request.parameters.*"]}


_expected_attributes = {
    "agent": TRANS_EVENT_AGENT_KEYS + REQ_PARAMS,
    "user": USER_ATTRS,
    "intrinsic": TRANS_EVENT_INTRINSICS,
}

_expected_absent_attributes = {"agent": ["wsgi.output.seconds"], "user": [], "intrinsic": []}


@validate_transaction_event_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_event_include_request_params(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {
    "browser_monitoring.attributes.enabled": True,
    "browser_monitoring.attributes.include": ["request.parameters.*"],
}

_expected_attributes = {"agent": REQ_PARAMS, "user": USER_ATTRS, "intrinsic": BROWSER_INTRINSIC_KEYS}

_expected_absent_attributes = {"agent": ABSENT_BROWSER_KEYS, "user": []}


@validate_browser_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_browser_include_request_params(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test include and exclude request parameters.

_override_settings = {
    "error_collector.attributes.include": ["request.parameters.*"],
    "error_collector.attributes.exclude": ["request.parameters." + URL_PARAM2],
}

_expected_attributes = {
    "agent": TRACE_ERROR_AGENT_KEYS + ["request.parameters." + URL_PARAM],
    "user": ERROR_USER_ATTRS,
    "intrinsic": ["trip_id"],
}

_expected_attributes_event = {
    "agent": TRACE_ERROR_AGENT_KEYS + ["request.parameters." + URL_PARAM],
    "user": ERROR_USER_ATTRS,
    "intrinsic": ERROR_EVENT_INTRINSICS,
}

_expected_absent_attributes = {"agent": ["request.parameters." + URL_PARAM2], "user": [], "intrinsic": []}


@validate_error_event_attributes(_expected_attributes_event, _expected_absent_attributes)
@validate_transaction_error_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_error_in_transaction_include_exclude(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {
    "transaction_tracer.attributes.include": ["request.parameters.*"],
    "transaction_tracer.attributes.exclude": ["request.parameters." + URL_PARAM2],
}

_expected_attributes = {
    "agent": TRACE_ERROR_AGENT_KEYS + ["request.parameters." + URL_PARAM],
    "user": USER_ATTRS,
    "intrinsic": ["trip_id"],
}


@validate_transaction_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_trace_include_exclude(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {
    "transaction_events.attributes.include": ["request.parameters.*"],
    "transaction_events.attributes.exclude": ["request.parameters." + URL_PARAM2],
}

_expected_attributes = {
    "agent": TRANS_EVENT_AGENT_KEYS + ["request.parameters." + URL_PARAM],
    "user": USER_ATTRS,
    "intrinsic": TRANS_EVENT_INTRINSICS,
}

_expected_absent_attributes = {"agent": ["request.parameters." + URL_PARAM2], "user": [], "intrinsic": []}


@validate_transaction_event_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_event_include_exclude(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {
    "browser_monitoring.attributes.enabled": True,
    "browser_monitoring.attributes.include": ["request.parameters.*"],
    "browser_monitoring.attributes.exclude": ["request.parameters." + URL_PARAM2],
}

_expected_attributes = {
    "agent": ["request.parameters." + URL_PARAM],
    "user": USER_ATTRS,
    "intrinsic": BROWSER_INTRINSIC_KEYS,
}

_expected_absent_attributes = {"agent": ABSENT_BROWSER_KEYS + ["request.parameters." + URL_PARAM2], "user": []}


@validate_browser_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_browser_include_exclude_request_params(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test exclude user attribute.

_override_settings = {"error_collector.attributes.exclude": ["puppies"]}

_expected_attributes = {"agent": TRACE_ERROR_AGENT_KEYS, "user": ["sunshine", "ohnoes"], "intrinsic": ["trip_id"]}

_expected_attributes_event = {
    "agent": TRACE_ERROR_AGENT_KEYS,
    "user": ["sunshine", "ohnoes"],
    "intrinsic": ERROR_EVENT_INTRINSICS,
}

_expected_absent_attributes = {"agent": REQ_PARAMS, "user": ["puppies"], "intrinsic": []}


@validate_error_event_attributes(_expected_attributes_event, _expected_absent_attributes)
@validate_transaction_error_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_error_in_transaction_exclude_user_attribute(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"transaction_tracer.attributes.exclude": ["puppies"]}

_expected_attributes = {"agent": TRACE_ERROR_AGENT_KEYS, "user": ["sunshine"], "intrinsic": ["trip_id"]}


@validate_transaction_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_trace_exclude_user_attribute(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"transaction_events.attributes.exclude": ["puppies"]}

_expected_attributes = {"agent": TRANS_EVENT_AGENT_KEYS, "user": ["sunshine"], "intrinsic": TRANS_EVENT_INTRINSICS}

_expected_absent_attributes = {"agent": REQ_PARAMS, "user": ["puppies"], "intrinsic": []}


@validate_transaction_event_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_event_exclude_user_attribute(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"span_events.attributes.exclude": ["puppies"]}


@override_application_settings(_override_settings)
@dt_enabled
@validate_span_events(expected_users=_expected_attributes["user"], unexpected_users=_expected_absent_attributes["user"])
def test_span_event_exclude_user_attribute(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {
    "browser_monitoring.attributes.enabled": True,
    "browser_monitoring.attributes.exclude": ["puppies"],
}

_expected_attributes = {"agent": [], "user": ["sunshine"], "intrinsic": BROWSER_INTRINSIC_KEYS}

_expected_absent_attributes = {"agent": ABSENT_BROWSER_KEYS + REQ_PARAMS, "user": ["puppies"]}


@validate_browser_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_browser_exclude_user_attribute(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test exclude agent attribute.

_override_settings = {"attributes.exclude": ["request.*"], "attributes.include": ["request.headers.*"]}

_expected_attributes = {
    "agent": ["response.status", "request.headers.contentType", "request.headers.contentLength"],
    "user": ERROR_USER_ATTRS,
    "intrinsic": ["trip_id"],
}

_expected_attributes_event = {
    "agent": ["response.status", "request.headers.contentType", "request.headers.contentLength"],
    "user": ERROR_USER_ATTRS,
    "intrinsic": ERROR_EVENT_INTRINSICS,
}

_expected_absent_attributes = {"agent": ["request.method", "request.uri"] + REQ_PARAMS, "user": [], "intrinsic": []}


@validate_error_event_attributes(_expected_attributes_event, _expected_absent_attributes)
@validate_transaction_error_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_error_in_transaction_exclude_agent_attribute(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_expected_attributes = {
    "agent": ["response.status", "request.headers.contentType", "request.headers.contentLength"],
    "user": USER_ATTRS,
    "intrinsic": ["trip_id"],
}


@validate_transaction_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_trace_exclude_agent_attribute(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_expected_attributes = {
    "agent": ["response.status", "request.headers.contentType", "request.headers.contentLength"],
    "user": USER_ATTRS,
    "intrinsic": TRANS_EVENT_INTRINSICS,
}

_expected_absent_attributes = {"agent": ["request.method", "request.uri"] + REQ_PARAMS, "user": [], "intrinsic": []}


@validate_transaction_event_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_event_exclude_agent_attribute(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"attributes.exclude": ["request.*"], "attributes.include": ["request.headers.*"]}

_expected_agent_attributes = ["response.status", "request.headers.contentType", "request.headers.contentLength"]

_expected_absent_agent_attributes = ["request.method", "request.uri"] + REQ_PARAMS


@override_application_settings(_override_settings)
@dt_enabled
@validate_span_events(expected_agents=_expected_agent_attributes, unexpected_agents=_expected_absent_agent_attributes)
def test_span_event_exclude_agent_attribute(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# The only agent attributes browser has are request parameters, which are
# tested in the request parameters test cases

# Test capture_params True.

_override_settings = {"capture_params": True}

_expected_attributes = {"agent": AGENT_KEYS_ALL, "user": ERROR_USER_ATTRS, "intrinsic": ["trip_id"]}

_expected_attributes_event = {"agent": AGENT_KEYS_ALL, "user": ERROR_USER_ATTRS, "intrinsic": ERROR_EVENT_INTRINSICS}


@validate_error_event_attributes(_expected_attributes_event)
@validate_transaction_error_trace_attributes(_expected_attributes)
@override_application_settings(_override_settings)
def test_error_in_transaction_deprecated_capture_params_true(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_expected_attributes = {"agent": AGENT_KEYS_ALL, "user": USER_ATTRS, "intrinsic": ["trip_id"]}


@validate_transaction_trace_attributes(_expected_attributes)
@override_application_settings(_override_settings)
def test_transaction_trace_deprecated_capture_params_true(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test capture_params should not affect transaction events or browser.

_expected_attributes = {"agent": TRANS_EVENT_AGENT_KEYS, "user": USER_ATTRS, "intrinsic": TRANS_EVENT_INTRINSICS}

_expected_absent_attributes = {"agent": ["wsgi.output.seconds"] + REQ_PARAMS, "user": [], "intrinsic": []}


@validate_transaction_event_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_event_deprecated_capture_params_true(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"browser_monitoring.attributes.enabled": True, "capture_params": True}

_expected_attributes = {"agent": [], "user": USER_ATTRS, "intrinsic": BROWSER_INTRINSIC_KEYS}

_expected_absent_attributes = {"agent": ABSENT_BROWSER_KEYS + REQ_PARAMS, "user": []}


@validate_browser_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_browser_deprecated_capture_params_true(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test capture_params False.

_override_settings = {"capture_params": False}

_expected_attributes = {"agent": TRACE_ERROR_AGENT_KEYS, "user": ERROR_USER_ATTRS, "intrinsic": ["trip_id"]}

_expected_attributes_event = {
    "agent": TRACE_ERROR_AGENT_KEYS,
    "user": ERROR_USER_ATTRS,
    "intrinsic": ERROR_EVENT_INTRINSICS,
}

_expected_absent_attributes = {"agent": REQ_PARAMS, "user": [], "intrinsic": []}


@validate_error_event_attributes(_expected_attributes_event, _expected_absent_attributes)
@validate_transaction_error_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_error_in_transaction_deprecated_capture_params_false(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_expected_attributes = {"agent": TRACE_ERROR_AGENT_KEYS, "user": USER_ATTRS, "intrinsic": ["trip_id"]}


@validate_transaction_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_trace_deprecated_capture_params_false(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test capture_params should not affect transaction events.

_expected_attributes = {"agent": TRANS_EVENT_AGENT_KEYS, "user": USER_ATTRS, "intrinsic": TRANS_EVENT_INTRINSICS}

_expected_absent_attributes = {"agent": ["wsgi.output.seconds"] + REQ_PARAMS, "user": [], "intrinsic": []}


@validate_transaction_event_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_event_deprecated_capture_params_false(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"browser_monitoring.attributes.enabled": True, "capture_params": False}

_expected_attributes = {"agent": [], "user": USER_ATTRS, "intrinsic": BROWSER_INTRINSIC_KEYS}

_expected_absent_attributes = {"agent": ABSENT_BROWSER_KEYS + REQ_PARAMS, "user": []}


@validate_browser_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_browser_deprecated_capture_params_false(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test attempt to exclude intrinsic.

_override_settings = {"error_collector.attributes.exclude": ["trip_id"]}

_expected_attributes = {"agent": TRACE_ERROR_AGENT_KEYS, "user": ERROR_USER_ATTRS, "intrinsic": ["trip_id"]}

_expected_attributes_event = {
    "agent": TRACE_ERROR_AGENT_KEYS,
    "user": ERROR_USER_ATTRS,
    "intrinsic": ERROR_EVENT_INTRINSICS,
}

_expected_absent_attributes = {"agent": REQ_PARAMS, "user": [], "intrinsic": []}


@validate_error_event_attributes(_expected_attributes_event, _expected_absent_attributes)
@validate_transaction_error_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_error_in_transaction_exclude_intrinsic(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"transaction_tracer.attributes.exclude": ["trip_id"]}

_expected_attributes = {"agent": TRACE_ERROR_AGENT_KEYS, "user": USER_ATTRS, "intrinsic": ["trip_id"]}


@validate_transaction_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_trace_exclude_intrinsic(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"transaction_events.attributes.exclude": ["name", "duration", "timestamp", "type"]}

_expected_attributes = {"agent": TRANS_EVENT_AGENT_KEYS, "user": USER_ATTRS, "intrinsic": TRANS_EVENT_INTRINSICS}

_expected_absent_attributes = {"agent": ["wsgi.output.seconds"] + REQ_PARAMS, "user": [], "intrinsic": []}


@validate_transaction_event_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_event_exclude_intrinsic(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {
    "browser_monitoring.attributes.enabled": True,
    "browser_monitoring.attributes.exclude": BROWSER_INTRINSIC_KEYS,
}

_expected_attributes = {"agent": [], "user": USER_ATTRS, "intrinsic": BROWSER_INTRINSIC_KEYS}

_expected_absent_attributes = {"agent": ABSENT_BROWSER_KEYS + REQ_PARAMS, "user": []}


@validate_browser_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_browser_exclude_intrinsic(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test attributes off.

_override_settings = {"error_collector.attributes.enabled": False}

_expected_attributes = {"user": [], "agent": [], "intrinsic": ["trip_id"]}

_expected_attributes_event = {"user": [], "agent": [], "intrinsic": ERROR_EVENT_INTRINSICS}

_expected_absent_attributes = {"agent": AGENT_KEYS_ALL, "user": ERROR_USER_ATTRS, "intrinsic": []}


@validate_error_event_attributes(_expected_attributes_event, _expected_absent_attributes)
@validate_transaction_error_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_error_in_transaction_attributes_disabled(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"transaction_tracer.attributes.enabled": False}


@validate_transaction_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_trace_attributes_disabled(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


_override_settings = {"transaction_events.attributes.enabled": False}

_expected_attributes = {"user": [], "agent": [], "intrinsic": TRANS_EVENT_INTRINSICS}

_expected_absent_attributes = {"agent": AGENT_KEYS_ALL, "user": USER_ATTRS, "intrinsic": []}


@validate_transaction_event_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_transaction_event_attributes_disabled(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test browser monitoring off by default.

_expected_attributes = {"agent": [], "user": [], "intrinsic": BROWSER_INTRINSIC_KEYS}

_expected_absent_attributes = {"agent": ABSENT_BROWSER_KEYS + REQ_PARAMS, "user": USER_ATTRS}


@validate_browser_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings({})
def test_browser_attributes_disabled(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test exclude error parameter.

_override_settings = {"error_collector.attributes.exclude": ERROR_PARAMS}

_expected_attributes = {"agent": TRACE_ERROR_AGENT_KEYS, "user": USER_ATTRS, "intrinsic": ["trip_id"]}

_expected_attributes_event = {"agent": TRACE_ERROR_AGENT_KEYS, "user": USER_ATTRS, "intrinsic": ERROR_EVENT_INTRINSICS}

_expected_absent_attributes = {"agent": REQ_PARAMS, "user": ERROR_PARAMS, "intrinsic": []}


@validate_error_event_attributes(_expected_attributes_event, _expected_absent_attributes)
@validate_transaction_error_trace_attributes(_expected_attributes, _expected_absent_attributes)
@override_application_settings(_override_settings)
def test_error_in_transaction_error_param_excluded(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test browser monitoring disabled.

_override_settings = {"browser_monitoring.enabled": False}

_expected_attributes = {"agent": TRANS_EVENT_AGENT_KEYS, "user": USER_ATTRS, "intrinsic": []}


@validate_transaction_trace_attributes(_expected_attributes)
@validate_transaction_event_attributes(_expected_attributes)
@override_application_settings(_override_settings)
def test_browser_monitoring_disabled(normal_application):
    normal_application.get(REQUEST_URL, headers=REQUEST_HEADERS)


# Test outside transaction (error traces and events only).

INTRSICS_NO_TRANS = ("type", "error.class", "error.message", "error.expected", "timestamp", "transactionName")


class OutsideWithParamsError(Exception):
    pass


OutsideWithParamsError.name = callable_name(OutsideWithParamsError)


_expected_attributes = {"user": ["test_key"], "agent": [], "intrinsic": []}

_expected_attributes_event = {"user": ["test_key"], "agent": [], "intrinsic": INTRSICS_NO_TRANS}


@reset_core_stats_engine()
@validate_error_event_attributes_outside_transaction(_expected_attributes_event)
@validate_error_trace_attributes_outside_transaction(OutsideWithParamsError.name, _expected_attributes)
def test_error_outside_transaction():

    try:
        raise OutsideWithParamsError("Error outside transaction")
    except OutsideWithParamsError:
        application_instance = application()
        application_instance.notice_error(attributes={"test_key": "test_value"})


class OutsideNoParamsError(Exception):
    pass


OutsideNoParamsError.name = callable_name(OutsideNoParamsError)

_override_settings = {"error_collector.attributes.exclude": ["test_key"]}

_expected_attributes = {}

_expected_attributes_event = {"user": [], "agent": [], "intrinsic": INTRSICS_NO_TRANS}

_expected_absent_attributes = {"user": ["test_key"], "agent": [], "intrinsic": []}


@reset_core_stats_engine()
@validate_error_event_attributes_outside_transaction(_expected_attributes_event, _expected_absent_attributes)
@validate_error_trace_attributes_outside_transaction(
    OutsideNoParamsError.name, _expected_attributes, _expected_absent_attributes
)
@override_application_settings(_override_settings)
def test_error_outside_transaction_excluded_user_param():

    try:
        raise OutsideNoParamsError("Error outside transaction")
    except OutsideNoParamsError:
        application_instance = application()
        application_instance.notice_error(attributes={"test_key": "test_value"})


# Test routing key agent attribute.

_required_agent_attributes = ["message.routingKey"]
_forgone_agent_attributes = []


@validate_attributes("agent", _required_agent_attributes, _forgone_agent_attributes)
@message_transaction(
    library="RabbitMQ", destination_type="Exchange", destination_name="x", routing_key="cat.eat.fishies"
)
def test_routing_key_agent_attribute():
    pass


_required_agent_attributes = []
_forgone_agent_attributes = ["message.routingKey"]


@validate_attributes("agent", _required_agent_attributes, _forgone_agent_attributes)
@message_transaction(library="RabbitMQ", destination_type="Exchange", destination_name="x")
def test_none_type_routing_key_agent_attribute():
    pass


_required_agent_attributes = ["enduser.id"]
_forgone_agent_attributes = []


@pytest.mark.parametrize(
    "input_user_id, reported_user_id, high_security",
    (
        ("1234", "1234", True),
        ("a" * 260, "a" * 255, False),
    ),
)
def test_enduser_id_attribute_api_valid_types(input_user_id, reported_user_id, high_security):
    @reset_core_stats_engine()
    @validate_error_trace_attributes(
        callable_name(ValueError), exact_attrs={"user": {}, "intrinsic": {}, "agent": {"enduser.id": reported_user_id}}
    )
    @validate_error_event_attributes(
        exact_attrs={"user": {}, "intrinsic": {}, "agent": {"enduser.id": reported_user_id}}
    )
    @validate_attributes("agent", _required_agent_attributes, _forgone_agent_attributes)
    @background_task()
    @override_application_settings({"high_security": high_security})
    def _test():
        set_user_id(input_user_id)

        try:
            raise ValueError()
        except Exception:
            notice_error()

    _test()


@pytest.mark.parametrize("input_user_id", (None, "", 123))
def test_enduser_id_attribute_api_invalid_types(input_user_id):
    @reset_core_stats_engine()
    @validate_attributes("agent", [], ["enduser.id"])
    @background_task()
    def _test():
        set_user_id(input_user_id)

        try:
            raise ValueError()
        except Exception:
            notice_error()

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import pandas
from testing_support.fixtures import reset_core_stats_engine
from testing_support.validators.validate_ml_event_count import validate_ml_event_count
from testing_support.validators.validate_ml_events import validate_ml_events

from newrelic.api.background_task import background_task
from newrelic.api.ml_model import wrap_mlmodel

try:
    from sklearn.tree._classes import BaseDecisionTree
except ImportError:
    from sklearn.tree.tree import BaseDecisionTree

_logger = logging.getLogger(__name__)


# Create custom model that isn't auto-instrumented to validate ml_model wrapper functionality
class CustomTestModel(BaseDecisionTree):
    def __init__(
        self,
        criterion="poisson",
        splitter="random",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features=None,
        random_state=0,
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        class_weight=None,
        ccp_alpha=0.0,
    ):
        super().__init__(
            criterion=criterion,
            splitter=splitter,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            class_weight=class_weight,
            random_state=random_state,
            min_impurity_decrease=min_impurity_decrease,
            ccp_alpha=ccp_alpha,
        )

    def fit(self, X, y, sample_weight=None, check_input=True):
        if hasattr(super(CustomTestModel, self), "_fit"):
            return self._fit(
                X,
                y,
                sample_weight=sample_weight,
                check_input=check_input,
            )
        else:
            return super(CustomTestModel, self).fit(
                X,
                y,
                sample_weight=sample_weight,
                check_input=check_input,
            )

    def predict(self, X, check_input=True):
        return super(CustomTestModel, self).predict(X, check_input=check_input)


int_list_recorded_custom_events = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "MyCustomModel",
            "model_version": "1.2.3",
            "feature.0": 1.0,
            "feature.1": 2.0,
            "label.0": "0.5",
            "new_relic_data_schema_version": 2,
        },
    ),
]


@reset_core_stats_engine()
def test_custom_model_int_list_no_features_and_labels():
    @validate_ml_event_count(count=1)
    @validate_ml_events(int_list_recorded_custom_events)
    @background_task()
    def _test():
        x_train = [[0, 0], [1, 1]]
        y_train = [0, 1]
        x_test = [[1.0, 2.0]]

        model = CustomTestModel().fit(x_train, y_train)
        wrap_mlmodel(model, name="MyCustomModel", version="1.2.3")

        labels = model.predict(x_test)

        return model

    _test()


int_list_recorded_custom_events_with_metadata = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "MyCustomModel",
            "model_version": "1.2.3",
            "feature.0": 1.0,
            "feature.1": 2.0,
            "label.0": "0.5",
            "new_relic_data_schema_version": 2,
            "metadata1": "value1",
            "metadata2": "value2",
        },
    ),
]


@reset_core_stats_engine()
def test_custom_model_int_list_with_metadata():
    @validate_ml_event_count(count=1)
    @validate_ml_events(int_list_recorded_custom_events_with_metadata)
    @background_task()
    def _test():
        x_train = [[0, 0], [1, 1]]
        y_train = [0, 1]
        x_test = [[1.0, 2.0]]

        model = CustomTestModel().fit(x_train, y_train)
        wrap_mlmodel(
            model,
            name="MyCustomModel",
            version="1.2.3",
            metadata={"metadata1": "value1", "metadata2": "value2"},
        )

        labels = model.predict(x_test)

        return model

    _test()


pandas_df_recorded_custom_events = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "PandasTestModel",
            "model_version": "1.5.0b1",
            "feature.feature1": 0,
            "feature.feature2": 0,
            "feature.feature3": 1,
            "label.label1": "0.5",
            "new_relic_data_schema_version": 2,
        },
    ),
]


@reset_core_stats_engine()
def test_wrapper_attrs_custom_model_pandas_df():
    @validate_ml_event_count(count=1)
    @validate_ml_events(pandas_df_recorded_custom_events)
    @background_task()
    def _test():
        x_train = pandas.DataFrame({"col1": [0, 1], "col2": [0, 1], "col3": [1, 2]}, dtype="category")
        y_train = [0, 1]
        x_test = pandas.DataFrame({"col1": [0], "col2": [0], "col3": [1]}, dtype="category")

        model = CustomTestModel(random_state=0).fit(x_train, y_train)
        wrap_mlmodel(
            model,
            name="PandasTestModel",
            version="1.5.0b1",
            feature_names=["feature1", "feature2", "feature3"],
            label_names=["label1"],
        )
        model.predict(x_test)
        return model

    _test()


pandas_df_recorded_builtin_events = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "MyDecisionTreeClassifier",
            "model_version": "1.5.0b1",
            "feature.feature1": 12,
            "feature.feature2": 14,
            "label.label1": "0",
            "new_relic_data_schema_version": 2,
        },
    ),
]


@reset_core_stats_engine()
def test_wrapper_attrs_builtin_model():
    @validate_ml_event_count(count=1)
    @validate_ml_events(pandas_df_recorded_builtin_events)
    @background_task()
    def _test():
        import sklearn.tree

        x_train = pandas.DataFrame({"col1": [0, 0], "col2": [1, 1]}, dtype="int")
        y_train = pandas.DataFrame({"label": [0, 1]}, dtype="int")
        x_test = pandas.DataFrame({"col1": [12], "col2": [14]}, dtype="int")

        clf = getattr(sklearn.tree, "DecisionTreeClassifier")(random_state=0)

        model = clf.fit(x_train, y_train)
        wrap_mlmodel(
            model,
            name="MyDecisionTreeClassifier",
            version="1.5.0b1",
            feature_names=["feature1", "feature2"],
            label_names=["label1"],
        )
        model.predict(x_test)

        return model

    _test()


pandas_df_mismatched_custom_events = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "MyDecisionTreeClassifier",
            "model_version": "1.5.0b1",
            "feature.col1": 12,
            "feature.col2": 14,
            "feature.col3": 16,
            "label.0": "1",
            "new_relic_data_schema_version": 2,
        },
    ),
]


@reset_core_stats_engine()
def test_wrapper_mismatched_features_and_labels_df():
    @validate_ml_event_count(count=1)
    @validate_ml_events(pandas_df_mismatched_custom_events)
    @background_task()
    def _test():
        import sklearn.tree

        x_train = pandas.DataFrame({"col1": [7, 8], "col2": [9, 10], "col3": [24, 25]}, dtype="int")
        y_train = pandas.DataFrame({"label": [0, 1]}, dtype="int")
        x_test = pandas.DataFrame({"col1": [12], "col2": [14], "col3": [16]}, dtype="int")

        clf = getattr(sklearn.tree, "DecisionTreeClassifier")(random_state=0)

        model = clf.fit(x_train, y_train)
        wrap_mlmodel(
            model,
            name="MyDecisionTreeClassifier",
            version="1.5.0b1",
            feature_names=["feature1", "feature2"],
            label_names=["label1", "label2"],
        )
        model.predict(x_test)
        return model

    _test()


numpy_str_mismatched_custom_events = [
    (
        {"type": "InferenceData"},
        {
            "inference_id": None,
            "prediction_id": None,
            "modelName": "MyDecisionTreeClassifier",
            "model_version": "0.0.1",
            "feature.0": "20",
            "feature.1": "21",
            "label.0": "21",
            "new_relic_data_schema_version": 2,
        },
    ),
]


@reset_core_stats_engine()
def test_wrapper_mismatched_features_and_labels_np_array():
    @validate_ml_events(numpy_str_mismatched_custom_events)
    @validate_ml_event_count(count=1)
    @background_task()
    def _test():
        import numpy as np
        import sklearn.tree

        x_train = np.array([[20, 20], [21, 21]], dtype="<U4")
        y_train = np.array([20, 21], dtype="<U4")
        x_test = np.array([[20, 21]], dtype="<U4")
        clf = getattr(sklearn.tree, "DecisionTreeClassifier")(random_state=0)

        model = clf.fit(x_train, y_train)
        wrap_mlmodel(model, name="MyDecisionTreeClassifier", version="0.0.1", feature_names=["feature1"])
        model.predict(x_test)

        return model

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import threading

from newrelic.packages.six.moves import BaseHTTPServer

# This defines an external server test apps can make requests to (instead of
# www.google.com for example). This provides 3 features:
#
# 1) This removes dependencies on external websites.
# 2) Provides a better mechanism for making an external call in a test app than
#    simple calling another endpoint the test app makes available because this
#    server will not be instrumented meaning we don't have to sort through
#    transactions to separate the ones created in the test app and the ones
#    created by an external call.
# 3) This app runs on a separate thread meaning it won't block the test app.


def simple_get(self):
    self.send_response(200)
    self.end_headers()
    self.wfile.write(b'external response')


class MockExternalHTTPServer(threading.Thread):
    # To use this class in a test one needs to start and stop this server
    # before and after making requests to the test app that makes the external
    # calls. For an example see:
    # ../framework_tornado_r3/test_async_application.py
    RESPONSE = b'external response'

    def __init__(self, handler=simple_get, port=None, *args, **kwargs):
        super(MockExternalHTTPServer, self).__init__(*args, **kwargs)
        self.daemon = True
        handler = type('ResponseHandler',
                (BaseHTTPServer.BaseHTTPRequestHandler, object,),
                {
                    'do_GET': handler,
                    'do_OPTIONS': handler,
                    'do_HEAD': handler,
                    'do_POST': handler,
                    'do_PUT': handler,
                    'do_PATCH': handler,
                    'do_DELETE': handler,
                })

        if port:
            self.httpd = BaseHTTPServer.HTTPServer(('localhost', port), handler)
            self.port = port
        else:
            # If port not set, try to bind to a port until successful
            retries = 5  # Set retry limit to prevent infinite loops
            self.port = None  # Initialize empty
            while not self.port and retries > 0:
                retries -= 1
                try:
                    # Obtain random open port
                    port = self.get_open_port()
                    # Attempt to bind to port
                    self.httpd = BaseHTTPServer.HTTPServer(('localhost', port), handler)
                    self.port = port
                except OSError as exc:
                    # Reraise errors other than port already in use
                    if "Address already in use" not in exc:
                        raise


    @staticmethod
    def get_open_port():
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("", 0))
        port = s.getsockname()[1]
        s.close()
        return port

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, tb):
        self.stop()

    def run(self):
        self.httpd.serve_forever()

    def stop(self):
        # Shutdowns the httpd server.
        self.httpd.shutdown()
        # Close the socket so we can reuse it.
        self.httpd.socket.close()
        self.join()


def incoming_headers_to_body_text(self):
    response = str(self.headers).encode('utf-8')
    self.send_response(200)
    self.end_headers()
    self.wfile.write(response)


class MockExternalHTTPHResponseHeadersServer(MockExternalHTTPServer):
    """
    MockExternalHTTPHResponseHeadersServer will send the incoming
    request headers back as the response.body, allowing us to validate
    httpclient request headers.

    """

    def __init__(self, handler=incoming_headers_to_body_text, port=None):
        super(MockExternalHTTPHResponseHeadersServer, self).__init__(
                handler=handler, port=port)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc

import grpc
import pytest

from testing_support.fixtures import collector_agent_registration_fixture, collector_available_fixture  # noqa: F401; pylint: disable=W0611
from testing_support.mock_external_grpc_server import MockExternalgRPCServer

import newrelic.packages.six as six


_default_settings = {
    "transaction_tracer.explain_threshold": 0.0,
    "transaction_tracer.transaction_threshold": 0.0,
    "transaction_tracer.stack_trace_threshold": 0.0,
    "debug.log_data_collector_payloads": True,
    "debug.record_transaction_failure": True,
}

collector_agent_registration = collector_agent_registration_fixture(
    app_name="Python Agent Test (framework_grpc)", default_settings=_default_settings
)


@pytest.fixture(scope="session")
def grpc_app_server():
    with MockExternalgRPCServer() as server:
        yield server, server.port


@pytest.fixture(scope="session")
def mock_grpc_server(grpc_app_server):
    from sample_application import (
        SampleApplicationServicer,
        add_SampleApplicationServicer_to_server,
    )

    server, port = grpc_app_server
    add_SampleApplicationServicer_to_server(SampleApplicationServicer(), server)
    return port


@pytest.fixture(scope="function", autouse=True)
def gc_garbage_empty():
    yield

    # Python 2 fails to collect objects with finalizers that participate in a reference cycle.
    # These assertions are made with that assumption in mind.
    # If there's a failure on py2, it's applicable to py3 as well.
    # If PY3 has a reference cycle (which it shouldn't), but PY2 does not, it will be GCed
    if six.PY2:
        # garbage collect until everything is reachable
        while gc.collect():
            pass

        from grpc._channel import _Rendezvous

        rendezvous_stored = sum(1 for o in gc.get_objects() if hasattr(o, "__class__") and isinstance(o, _Rendezvous))

        assert rendezvous_stored == 0

        # make sure that even python knows there isn't any garbage remaining
        assert not gc.garbage


@pytest.fixture(scope="session")
def stub(stub_and_channel):
    return stub_and_channel[0]


@pytest.fixture(scope="session")
def stub_and_channel(mock_grpc_server):
    port = mock_grpc_server
    stub, channel = create_stub_and_channel(port)
    with channel:
        yield stub, channel


def create_stub_and_channel(port):
    from sample_application import SampleApplicationStub

    channel = grpc.insecure_channel("localhost:%s" % port)
    stub = SampleApplicationStub(channel)
    return stub, channel



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from framework_graphql.test_application_async import (
    error_middleware_async,
    example_middleware_async,
)
from testing_support.fixtures import dt_enabled, override_application_settings
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_count import (
    validate_transaction_count,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.background_task import background_task
from newrelic.common.object_names import callable_name
from newrelic.common.package_version_utils import get_package_version


graphql_version = get_package_version("graphql-core")

def conditional_decorator(decorator, condition):
    def _conditional_decorator(func):
        if not condition:
            return func
        return decorator(func)

    return _conditional_decorator


def to_graphql_source(query):
    def delay_import():
        try:
            from graphql import Source
        except ImportError:
            # Fallback if Source is not implemented
            return query

        return Source(query)

    return delay_import


def example_middleware(next, root, info, **args):
    return_value = next(root, info, **args)
    return return_value


def error_middleware(next, root, info, **args):
    raise RuntimeError("Runtime Error!")


def test_no_harm_no_transaction(target_application):
    framework, version, target_application, is_bg, schema_type, extra_spans = target_application

    def _test():
        response = target_application("{ __schema { types { name } } }")

    _test()


example_middleware = [example_middleware]
error_middleware = [error_middleware]

example_middleware.append(example_middleware_async)
error_middleware.append(error_middleware_async)

_runtime_error_name = callable_name(RuntimeError)
_test_runtime_error = [(_runtime_error_name, "Runtime Error!")]


def _graphql_base_rollup_metrics(framework, version, background_task=True):
    graphql_version = get_package_version("graphql-core")

    metrics = [
        ("Python/Framework/GraphQL/%s" % graphql_version, 1),
        ("GraphQL/all", 1),
        ("GraphQL/%s/all" % framework, 1),
    ]
    if background_task:
        metrics.extend(
            [
                ("GraphQL/allOther", 1),
                ("GraphQL/%s/allOther" % framework, 1),
            ]
        )
    else:
        metrics.extend(
            [
                ("GraphQL/allWeb", 1),
                ("GraphQL/%s/allWeb" % framework, 1),
            ]
        )

    if framework != "GraphQL":
        metrics.append(("Python/Framework/%s/%s" % (framework, version), 1))

    return metrics


def test_basic(target_application):
    framework, version, target_application, is_bg, schema_type, extra_spans = target_application

    @validate_transaction_metrics(
        "query/<anonymous>/hello",
        "GraphQL",
        rollup_metrics=_graphql_base_rollup_metrics(framework, version, is_bg),
        background_task=is_bg,
    )
    @conditional_decorator(background_task(), is_bg)
    def _test():
        response = target_application("{ hello }")
        assert response["hello"] == "Hello!"

    _test()


@dt_enabled
def test_query_and_mutation(target_application):
    framework, version, target_application, is_bg, schema_type, extra_spans = target_application

    mutation_path = "storage_add" if framework != "Graphene" else "storage_add.string"
    type_annotation = "!" if framework == "Strawberry" else ""

    _test_mutation_scoped_metrics = [
        ("GraphQL/resolve/%s/storage_add" % framework, 1),
        ("GraphQL/operation/%s/mutation/<anonymous>/%s" % (framework, mutation_path), 1),
    ]
    _test_query_scoped_metrics = [
        ("GraphQL/resolve/%s/storage" % framework, 1),
        ("GraphQL/operation/%s/query/<anonymous>/storage" % framework, 1),
    ]

    _expected_mutation_operation_attributes = {
        "graphql.operation.type": "mutation",
        "graphql.operation.name": "<anonymous>",
    }
    _expected_mutation_resolver_attributes = {
        "graphql.field.name": "storage_add",
        "graphql.field.parentType": "Mutation",
        "graphql.field.path": "storage_add",
        "graphql.field.returnType": ("String" if framework != "Graphene" else "StorageAdd") + type_annotation,
    }
    _expected_query_operation_attributes = {
        "graphql.operation.type": "query",
        "graphql.operation.name": "<anonymous>",
    }
    _expected_query_resolver_attributes = {
        "graphql.field.name": "storage",
        "graphql.field.parentType": "Query",
        "graphql.field.path": "storage",
        "graphql.field.returnType": "[String%s]%s" % (type_annotation, type_annotation),
    }

    @validate_code_level_metrics(
        "framework_%s._target_schema_%s" % (framework.lower(), schema_type), "resolve_storage_add"
    )
    @validate_span_events(exact_agents=_expected_mutation_operation_attributes)
    @validate_span_events(exact_agents=_expected_mutation_resolver_attributes)
    @validate_transaction_metrics(
        "mutation/<anonymous>/%s" % mutation_path,
        "GraphQL",
        scoped_metrics=_test_mutation_scoped_metrics,
        rollup_metrics=_test_mutation_scoped_metrics + _graphql_base_rollup_metrics(framework, version, is_bg),
        background_task=is_bg,
    )
    @conditional_decorator(background_task(), is_bg)
    def _mutation():
        if framework == "Graphene":
            query = 'mutation { storage_add(string: "abc") { string } }'
        else:
            query = 'mutation { storage_add(string: "abc") }'
        response = target_application(query)
        assert response["storage_add"] == "abc" or response["storage_add"]["string"] == "abc"

    @validate_code_level_metrics("framework_%s._target_schema_%s" % (framework.lower(), schema_type), "resolve_storage")
    @validate_span_events(exact_agents=_expected_query_operation_attributes)
    @validate_span_events(exact_agents=_expected_query_resolver_attributes)
    @validate_transaction_metrics(
        "query/<anonymous>/storage",
        "GraphQL",
        scoped_metrics=_test_query_scoped_metrics,
        rollup_metrics=_test_query_scoped_metrics + _graphql_base_rollup_metrics(framework, version, is_bg),
        background_task=is_bg,
    )
    @conditional_decorator(background_task(), is_bg)
    def _query():
        response = target_application("query { storage }")
        assert response["storage"] == ["abc"]

    _mutation()
    _query()


@pytest.mark.parametrize("middleware", example_middleware)
@dt_enabled
def test_middleware(target_application, middleware):
    framework, version, target_application, is_bg, schema_type, extra_spans = target_application

    name = "%s:%s" % (middleware.__module__, middleware.__name__)
    if "async" in name:
        if schema_type != "async":
            pytest.skip("Async middleware not supported in sync applications.")

    _test_middleware_metrics = [
        ("GraphQL/operation/%s/query/<anonymous>/hello" % framework, 1),
        ("GraphQL/resolve/%s/hello" % framework, 1),
        ("Function/%s" % name, 1),
    ]

    # Span count 5: Transaction, Operation, Middleware, and 1 Resolver and Resolver Function
    span_count = 5 + extra_spans

    @validate_code_level_metrics(*name.split(":"))
    @validate_code_level_metrics("framework_%s._target_schema_%s" % (framework.lower(), schema_type), "resolve_hello")
    @validate_span_events(count=span_count)
    @validate_transaction_metrics(
        "query/<anonymous>/hello",
        "GraphQL",
        scoped_metrics=_test_middleware_metrics,
        rollup_metrics=_test_middleware_metrics + _graphql_base_rollup_metrics(framework, version, is_bg),
        background_task=is_bg,
    )
    @conditional_decorator(background_task(), is_bg)
    def _test():
        response = target_application("{ hello }", middleware=[middleware])
        assert response["hello"] == "Hello!"

    _test()


@pytest.mark.parametrize("middleware", error_middleware)
@dt_enabled
def test_exception_in_middleware(target_application, middleware):
    framework, version, target_application, is_bg, schema_type, extra_spans = target_application
    query = "query MyQuery { error_middleware }"
    field = "error_middleware"

    name = "%s:%s" % (middleware.__module__, middleware.__name__)
    if "async" in name:
        if schema_type != "async":
            pytest.skip("Async middleware not supported in sync applications.")

    # Metrics
    _test_exception_scoped_metrics = [
        ("GraphQL/operation/%s/query/MyQuery/%s" % (framework, field), 1),
        ("GraphQL/resolve/%s/%s" % (framework, field), 1),
        ("Function/%s" % name, 1),
    ]
    _test_exception_rollup_metrics = [
        ("Errors/all", 1),
        ("Errors/all%s" % ("Other" if is_bg else "Web"), 1),
        ("Errors/%sTransaction/GraphQL/%s" % ("Other" if is_bg else "Web", name), 1),
    ] + _test_exception_scoped_metrics

    # Attributes
    _expected_exception_resolver_attributes = {
        "graphql.field.name": field,
        "graphql.field.parentType": "Query",
        "graphql.field.path": field,
        "graphql.field.returnType": "String",
    }
    _expected_exception_operation_attributes = {
        "graphql.operation.type": "query",
        "graphql.operation.name": "MyQuery",
        "graphql.operation.query": query,
    }

    @validate_transaction_metrics(
        name,
        "GraphQL",
        scoped_metrics=_test_exception_scoped_metrics,
        rollup_metrics=_test_exception_rollup_metrics + _graphql_base_rollup_metrics(framework, version, is_bg),
        background_task=is_bg,
    )
    @validate_span_events(exact_agents=_expected_exception_operation_attributes)
    @validate_span_events(exact_agents=_expected_exception_resolver_attributes)
    @validate_transaction_errors(errors=_test_runtime_error)
    @conditional_decorator(background_task(), is_bg)
    def _test():
        response = target_application(query, middleware=[middleware])

    _test()


@pytest.mark.parametrize("field", ("error", "error_non_null"))
@dt_enabled
def test_exception_in_resolver(target_application, field):
    framework, version, target_application, is_bg, schema_type, extra_spans = target_application
    query = "query MyQuery { %s }" % field

    txn_name = "framework_%s._target_schema_%s:resolve_error" % (framework.lower(), schema_type)

    # Metrics
    _test_exception_scoped_metrics = [
        ("GraphQL/operation/%s/query/MyQuery/%s" % (framework, field), 1),
        ("GraphQL/resolve/%s/%s" % (framework, field), 1),
    ]
    _test_exception_rollup_metrics = [
        ("Errors/all", 1),
        ("Errors/all%s" % ("Other" if is_bg else "Web"), 1),
        ("Errors/%sTransaction/GraphQL/%s" % ("Other" if is_bg else "Web", txn_name), 1),
    ] + _test_exception_scoped_metrics

    # Attributes
    _expected_exception_resolver_attributes = {
        "graphql.field.name": field,
        "graphql.field.parentType": "Query",
        "graphql.field.path": field,
        "graphql.field.returnType": "String!" if "non_null" in field else "String",
    }
    _expected_exception_operation_attributes = {
        "graphql.operation.type": "query",
        "graphql.operation.name": "MyQuery",
        "graphql.operation.query": query,
    }

    @validate_transaction_metrics(
        txn_name,
        "GraphQL",
        scoped_metrics=_test_exception_scoped_metrics,
        rollup_metrics=_test_exception_rollup_metrics + _graphql_base_rollup_metrics(framework, version, is_bg),
        background_task=is_bg,
    )
    @validate_span_events(exact_agents=_expected_exception_operation_attributes)
    @validate_span_events(exact_agents=_expected_exception_resolver_attributes)
    @validate_transaction_errors(errors=_test_runtime_error)
    @conditional_decorator(background_task(), is_bg)
    def _test():
        response = target_application(query)

    _test()


@dt_enabled
@pytest.mark.parametrize(
    "query,exc_class",
    [
        ("query MyQuery { error_missing_field }", "GraphQLError"),
        ("{ syntax_error ", "graphql.error.syntax_error:GraphQLSyntaxError"),
    ],
)
def test_exception_in_validation(target_application, query, exc_class):
    framework, version, target_application, is_bg, schema_type, extra_spans = target_application
    if "syntax" in query:
        txn_name = "graphql.language.parser:parse"
    else:
        txn_name = "graphql.validation.validate:validate"

    # Import path differs between versions
    if exc_class == "GraphQLError":
        from graphql.error import GraphQLError

        exc_class = callable_name(GraphQLError)

    _test_exception_scoped_metrics = [
        ("GraphQL/operation/%s/<unknown>/<anonymous>/<unknown>" % framework, 1),
    ]
    _test_exception_rollup_metrics = [
        ("Errors/all", 1),
        ("Errors/all%s" % ("Other" if is_bg else "Web"), 1),
        ("Errors/%sTransaction/GraphQL/%s" % ("Other" if is_bg else "Web", txn_name), 1),
    ] + _test_exception_scoped_metrics

    # Attributes
    _expected_exception_operation_attributes = {
        "graphql.operation.type": "<unknown>",
        "graphql.operation.name": "<anonymous>",
        "graphql.operation.query": query,
    }

    @validate_transaction_metrics(
        txn_name,
        "GraphQL",
        scoped_metrics=_test_exception_scoped_metrics,
        rollup_metrics=_test_exception_rollup_metrics + _graphql_base_rollup_metrics(framework, version, is_bg),
        background_task=is_bg,
    )
    @validate_span_events(exact_agents=_expected_exception_operation_attributes)
    @validate_transaction_errors(errors=[exc_class])
    @conditional_decorator(background_task(), is_bg)
    def _test():
        response = target_application(query)

    _test()


@dt_enabled
def test_operation_metrics_and_attrs(target_application):
    framework, version, target_application, is_bg, schema_type, extra_spans = target_application
    operation_metrics = [("GraphQL/operation/%s/query/MyQuery/library" % framework, 1)]
    operation_attrs = {
        "graphql.operation.type": "query",
        "graphql.operation.name": "MyQuery",
    }

    # Span count 16: Transaction, Operation, and 7 Resolvers and Resolver functions
    # library, library.name, library.book
    # library.book.name and library.book.id for each book resolved (in this case 2)
    span_count = 16 + extra_spans  # WSGI may add 4 spans, other frameworks may add other amounts

    @validate_transaction_metrics(
        "query/MyQuery/library",
        "GraphQL",
        scoped_metrics=operation_metrics,
        rollup_metrics=operation_metrics + _graphql_base_rollup_metrics(framework, version, is_bg),
        background_task=is_bg,
    )
    @validate_span_events(count=span_count)
    @validate_span_events(exact_agents=operation_attrs)
    @conditional_decorator(background_task(), is_bg)
    def _test():
        response = target_application("query MyQuery { library(index: 0) { branch, book { id, name } } }")

    _test()


@dt_enabled
def test_field_resolver_metrics_and_attrs(target_application):
    framework, version, target_application, is_bg, schema_type, extra_spans = target_application
    field_resolver_metrics = [("GraphQL/resolve/%s/hello" % framework, 1)]

    type_annotation = "!" if framework == "Strawberry" else ""
    graphql_attrs = {
        "graphql.field.name": "hello",
        "graphql.field.parentType": "Query",
        "graphql.field.path": "hello",
        "graphql.field.returnType": "String" + type_annotation,
    }

    # Span count 4: Transaction, Operation, and 1 Resolver and Resolver function
    span_count = 4 + extra_spans  # WSGI may add 4 spans, other frameworks may add other amounts

    @validate_transaction_metrics(
        "query/<anonymous>/hello",
        "GraphQL",
        scoped_metrics=field_resolver_metrics,
        rollup_metrics=field_resolver_metrics + _graphql_base_rollup_metrics(framework, version, is_bg),
        background_task=is_bg,
    )
    @validate_span_events(count=span_count)
    @validate_span_events(exact_agents=graphql_attrs)
    @conditional_decorator(background_task(), is_bg)
    def _test():
        response = target_application("{ hello }")
        assert response["hello"] == "Hello!"

    _test()


_test_queries = [
    ("{ hello }", "{ hello }"),  # Basic query extraction
    ("{ error }", "{ error }"),  # Extract query on field error
    (to_graphql_source("{ hello }"), "{ hello }"),  # Extract query from Source objects
    ("{ library(index: 0) { branch } }", "{ library(index: ?) { branch } }"),  # Integers
    ('{ echo(echo: "123") }', "{ echo(echo: ?) }"),  # Strings with numerics
    ('{ echo(echo: "test") }', "{ echo(echo: ?) }"),  # Strings
    ('{ TestEcho: echo(echo: "test") }', "{ TestEcho: echo(echo: ?) }"),  # Aliases
    ('{ TestEcho: echo(echo: "test") }', "{ TestEcho: echo(echo: ?) }"),  # Variables
    (  # Fragments
        '{ ...MyFragment } fragment MyFragment on Query { echo(echo: "test") }',
        "{ ...MyFragment } fragment MyFragment on Query { echo(echo: ?) }",
    ),
]


@dt_enabled
@pytest.mark.parametrize("query,obfuscated", _test_queries)
def test_query_obfuscation(target_application, query, obfuscated):
    framework, version, target_application, is_bg, schema_type, extra_spans = target_application
    graphql_attrs = {"graphql.operation.query": obfuscated}

    if callable(query):
        if framework != "GraphQL":
            pytest.skip("Source query objects not tested outside of graphql-core")
        query = query()

    @validate_span_events(exact_agents=graphql_attrs)
    @conditional_decorator(background_task(), is_bg)
    def _test():
        response = target_application(query)

    _test()


_test_queries = [
    ("{ hello }", "/hello"),  # Basic query
    ("{ error }", "/error"),  # Extract deepest path on field error
    ('{ echo(echo: "test") }', "/echo"),  # Fields with arguments
    (
        "{ library(index: 0) { branch, book { isbn branch } } }",
        "/library",
    ),  # Complex Example, 1 level
    (
        "{ library(index: 0) { book { author { first_name }} } }",
        "/library.book.author.first_name",
    ),  # Complex Example, 2 levels
    ("{ library(index: 0) { id, book { name } } }", "/library.book.name"),  # Filtering
    ('{ TestEcho: echo(echo: "test") }', "/echo"),  # Aliases
    (
        '{ search(contains: "A") { __typename ... on Book { name } } }',
        "/search<Book>.name",
    ),  # InlineFragment
    (
        '{ hello echo(echo: "test") }',
        "",
    ),  # Multiple root selections. (need to decide on final behavior)
    # FragmentSpread
    (
        "{ library(index: 0) { book { ...MyFragment } } } fragment MyFragment on Book { name id }",  # Fragment filtering
        "/library.book.name",
    ),
    (
        "{ library(index: 0) { book { ...MyFragment } } } fragment MyFragment on Book { author { first_name } }",
        "/library.book.author.first_name",
    ),
    (
        "{ library(index: 0) { book { ...MyFragment } magazine { ...MagFragment } } } fragment MyFragment on Book { author { first_name } } fragment MagFragment on Magazine { name }",
        "/library",
    ),
]


@dt_enabled
@pytest.mark.parametrize("query,expected_path", _test_queries)
def test_deepest_unique_path(target_application, query, expected_path):
    framework, version, target_application, is_bg, schema_type, extra_spans = target_application
    if expected_path == "/error":
        txn_name = "framework_%s._target_schema_%s:resolve_error" % (framework.lower(), schema_type)
    else:
        txn_name = "query/<anonymous>%s" % expected_path

    @validate_transaction_metrics(
        txn_name,
        "GraphQL",
        background_task=is_bg,
    )
    @conditional_decorator(background_task(), is_bg)
    def _test():
        response = target_application(query)

    _test()


@pytest.mark.parametrize("capture_introspection_setting", (True, False))
def test_introspection_transactions(target_application, capture_introspection_setting):
    framework, version, target_application, is_bg, schema_type, extra_spans = target_application
    txn_ct = 1 if capture_introspection_setting else 0

    @override_application_settings(
        {"instrumentation.graphql.capture_introspection_queries": capture_introspection_setting}
    )
    @validate_transaction_count(txn_ct)
    @background_task()
    def _test():
        response = target_application("{ __schema { types { name } } }")

    _test()



# -*- coding: utf-8 -*-
# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
import time

import pytest
import webtest
from testing_support.fixtures import validate_attributes
from testing_support.sample_applications import simple_app, simple_app_raw
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.api.application import application_instance
from newrelic.api.web_transaction import WebTransaction
from newrelic.api.wsgi_application import wsgi_application
from newrelic.packages import six

application = webtest.TestApp(simple_app)


# TODO: WSGI metrics must not be generated for a WebTransaction
METRICS = (
    ("Python/WSGI/Input/Bytes", None),
    ("Python/WSGI/Input/Time", None),
    ("Python/WSGI/Input/Calls/read", None),
    ("Python/WSGI/Input/Calls/readline", None),
    ("Python/WSGI/Input/Calls/readlines", None),
    ("Python/WSGI/Output/Bytes", None),
    ("Python/WSGI/Output/Time", None),
    ("Python/WSGI/Output/Calls/yield", None),
    ("Python/WSGI/Output/Calls/write", None),
)


# Test for presence of framework and dispatcher info based on whether framework is specified
@validate_transaction_metrics(
    name="test", custom_metrics=[("Python/Framework/framework/v1", 1), ("Python/Dispatcher/dispatcher/v1.0.0", 1)]
)
def test_dispatcher_and_framework_metrics():
    inner_wsgi_decorator = wsgi_application(
        name="test", framework=("framework", "v1"), dispatcher=("dispatcher", "v1.0.0")
    )
    decorated_application = inner_wsgi_decorator(simple_app_raw)

    application = webtest.TestApp(decorated_application)
    application.get("/")


# Test for presence of framework and dispatcher info under existing transaction
@validate_transaction_metrics(
    name="test", custom_metrics=[("Python/Framework/framework/v1", 1), ("Python/Dispatcher/dispatcher/v1.0.0", 1)]
)
def test_double_wrapped_dispatcher_and_framework_metrics():
    inner_wsgi_decorator = wsgi_application(
        name="test", framework=("framework", "v1"), dispatcher=("dispatcher", "v1.0.0")
    )
    decorated_application = inner_wsgi_decorator(simple_app_raw)

    outer_wsgi_decorator = wsgi_application(name="double_wrapped")
    double_decorated_application = outer_wsgi_decorator(decorated_application)

    application = webtest.TestApp(double_decorated_application)
    application.get("/")


# TODO: Add rollup_metrics=METRICS
@validate_transaction_metrics("test_base_web_transaction", group="Test")
@validate_attributes(
    "agent",
    [
        "request.headers.accept",
        "request.headers.contentLength",
        "request.headers.contentType",
        "request.headers.host",
        "request.headers.referer",
        "request.headers.userAgent",
        "request.method",
        "request.uri",
        "response.status",
        "response.headers.contentLength",
        "response.headers.contentType",
        "request.parameters.foo",
        "request.parameters.boo",
        "webfrontend.queue.seconds",
    ],
)
@pytest.mark.parametrize("use_bytes", (True, False))
def test_base_web_transaction(use_bytes):
    application = application_instance()

    request_headers = {
        "Accept": "text/plain",
        "Content-Length": "0",
        "Content-Type": "text/plain",
        "Host": "localhost",
        "Referer": "http://example.com?q=1&boat=⛵",
        "User-Agent": "potato",
        "X-Request-Start": str(time.time() - 0.2),
        "newRelic": "invalid",
    }

    if use_bytes:
        byte_headers = {}

        for name, value in request_headers.items():
            name = name.encode("utf-8")
            try:
                value = value.encode("utf-8")
            except UnicodeDecodeError:
                assert six.PY2
            byte_headers[name] = value

        request_headers = byte_headers

    transaction = WebTransaction(
        application,
        "test_base_web_transaction",
        group="Test",
        scheme="http",
        host="localhost",
        port=8000,
        request_method="HEAD",
        request_path="/foobar",
        query_string="foo=bar&boo=baz",
        headers=request_headers.items(),
    )

    if use_bytes:
        response_headers = ((b"Content-Length", b"0"), (b"Content-Type", b"text/plain"))
    else:
        response_headers = (("Content-Length", "0"), ("Content-Type", "text/plain"))

    with transaction:
        transaction.process_response(200, response_headers)


@pytest.fixture()
def validate_no_garbage():
    yield

    # garbage collect until everything is reachable
    while gc.collect():
        pass

    assert not gc.garbage


@validate_transaction_metrics(
    name="",
    group="Uri",
)
def test_wsgi_app_memory(validate_no_garbage):
    application.get("/")



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

''' The purpose of these tests is to confirm that using a non-standard
connection pool that does not have a `connection_kwargs` attribute
will not result in an error.
'''

import pytest
import aredis

from newrelic.api.background_task import background_task

from testing_support.fixture.event_loop import event_loop as loop
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.db_settings import redis_settings
from testing_support.util import instance_hostname

DB_SETTINGS = redis_settings()[0]
REDIS_PY_VERSION = aredis.VERSION


class FakeConnectionPool(object):
    """Connection Pool without connection_kwargs attribute."""

    def __init__(self, connection):
        self.connection = connection

    def get_connection(self, name=None, *keys, **options):
        return self.connection

    def release(self, connection):
        self.connection.disconnect()

# Settings

_enable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': True,
}
_disable_instance_settings = {
    'datastore_tracer.instance_reporting.enabled': False,
}

# Metrics

# We don't record instance metrics when using redis blaster,
# so we just check for base metrics.

_base_scoped_metrics = (
        ('Datastore/operation/Redis/get', 1),
        ('Datastore/operation/Redis/set', 1),
        ('Datastore/operation/Redis/client_list', 1),
)

_base_rollup_metrics = (
        ('Datastore/all', 3),
        ('Datastore/allOther', 3),
        ('Datastore/Redis/all', 3),
        ('Datastore/Redis/allOther', 3),
        ('Datastore/operation/Redis/get', 1),
        ('Datastore/operation/Redis/set', 1),
        ('Datastore/operation/Redis/client_list', 1),
)

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

_host = instance_hostname(DB_SETTINGS['host'])
_port = DB_SETTINGS['port']

_instance_metric_name = 'Datastore/instance/Redis/%s/%s' % (_host, _port)

_enable_rollup_metrics.append(
        (_instance_metric_name, 3)
)

_disable_rollup_metrics.append(
        (_instance_metric_name, None)
)

# Operations

async def exercise_redis(client):
    await client.set('key', 'value')
    await client.get('key')
    await client.execute_command('CLIENT', 'LIST', parse='LIST')

# Tests

@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
        'test_custom_conn_pool:test_fake_conn_pool_enable_instance',
        scoped_metrics=_enable_scoped_metrics,
        rollup_metrics=_enable_rollup_metrics,
        background_task=True)
@background_task()
def test_fake_conn_pool_enable_instance(loop):
    client = aredis.StrictRedis(host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port'], db=0)

    # Get a real connection

    conn = client.connection_pool.get_connection('GET')

    # Replace the original connection pool with one that doesn't
    # have the `connection_kwargs` attribute.

    fake_pool = FakeConnectionPool(conn)
    client.connection_pool = fake_pool
    assert not hasattr(client.connection_pool, 'connection_kwargs')

    loop.run_until_complete(exercise_redis(client))

@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
        'test_custom_conn_pool:test_fake_conn_pool_disable_instance',
        scoped_metrics=_disable_scoped_metrics,
        rollup_metrics=_disable_rollup_metrics,
        background_task=True)
@background_task()
def test_fake_conn_pool_disable_instance(loop):
    client = aredis.StrictRedis(host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port'], db=0)

    # Get a real connection

    conn = client.connection_pool.get_connection('GET')

    # Replace the original connection pool with one that doesn't
    # have the `connection_kwargs` attribute.

    fake_pool = FakeConnectionPool(conn)
    client.connection_pool = fake_pool
    assert not hasattr(client.connection_pool, 'connection_kwargs')

    loop.run_until_complete(exercise_redis(client))



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import django
import pytest
import webtest
from testing_support.fixtures import function_not_called, override_generic_settings
from testing_support.validators.validate_code_level_metrics import (
    validate_code_level_metrics,
)
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)

from newrelic.core.config import global_settings
from newrelic.packages import six

DJANGO_VERSION = tuple(map(int, django.get_version().split(".")[:2]))

@pytest.fixture(scope="module")
def target_application():
    from wsgi import application

    test_application = webtest.TestApp(application)
    return test_application


if DJANGO_VERSION >= (1, 10):
    url_module_path = "django.urls.resolvers"

    # Django 1.10 new style middleware removed individual process_* methods.
    # All middleware in Django 1.10+ is called through the __call__ methods on
    # middlwares.
    process_request_method = ""
    process_view_method = ""
    process_response_method = ""
else:
    url_module_path = "django.core.urlresolvers"
    process_request_method = ".process_request"
    process_view_method = ".process_view"
    process_response_method = ".process_response"

if DJANGO_VERSION >= (2, 0):
    url_resolver_cls = "URLResolver"
else:
    url_resolver_cls = "RegexURLResolver"

_scoped_metrics = [
    ("Function/django.core.handlers.wsgi:WSGIHandler.__call__", 1),
    ("Python/WSGI/Application", 1),
    ("Python/WSGI/Response", 1),
    ("Python/WSGI/Finalize", 1),
    (("Function/django.middleware.common:CommonMiddleware%s" % process_request_method), 1),
    (("Function/django.contrib.sessions.middleware:SessionMiddleware%s" % process_request_method), 1),
    (("Function/django.contrib.auth.middleware:AuthenticationMiddleware%s" % process_request_method), 1),
    (("Function/django.contrib.messages.middleware:MessageMiddleware%s" % process_request_method), 1),
    (("Function/%s:%s.resolve" % (url_module_path, url_resolver_cls)), 1),
    (("Function/django.middleware.csrf:CsrfViewMiddleware%s" % process_view_method), 1),
    (("Function/django.contrib.messages.middleware:MessageMiddleware%s" % process_response_method), 1),
    (("Function/django.middleware.csrf:CsrfViewMiddleware%s" % process_response_method), 1),
    (("Function/django.contrib.sessions.middleware:SessionMiddleware%s" % process_response_method), 1),
    (("Function/django.middleware.common:CommonMiddleware%s" % process_response_method), 1),
]

_test_application_index_scoped_metrics = list(_scoped_metrics)
_test_application_index_scoped_metrics.append(("Function/views:index", 1))

if DJANGO_VERSION >= (1, 5):
    _test_application_index_scoped_metrics.extend([("Function/django.http.response:HttpResponse.close", 1)])


@validate_transaction_errors(errors=[])
@validate_transaction_metrics("views:index", scoped_metrics=_test_application_index_scoped_metrics)
@validate_code_level_metrics("views", "index")
def test_application_index(target_application):
    response = target_application.get("")
    response.mustcontain("INDEX RESPONSE")


_test_application_view_scoped_metrics = list(_scoped_metrics)
_test_application_view_scoped_metrics.append(("Function/urls:View.get", 1))

if DJANGO_VERSION >= (1, 5):
    _test_application_view_scoped_metrics.extend([("Function/rest_framework.response:Response.close", 1)])


@validate_transaction_errors(errors=[])
@validate_transaction_metrics("urls:View.get", scoped_metrics=_test_application_view_scoped_metrics)
@validate_code_level_metrics("urls.View", "get")
def test_application_view(target_application):
    response = target_application.get("/view/")
    assert response.status_int == 200
    response.mustcontain("restframework view response")


_test_application_view_error_scoped_metrics = list(_scoped_metrics)
_test_application_view_error_scoped_metrics.append(("Function/urls:ViewError.get", 1))


@validate_transaction_errors(errors=["urls:Error"])
@validate_transaction_metrics("urls:ViewError.get", scoped_metrics=_test_application_view_error_scoped_metrics)
@validate_code_level_metrics("urls.ViewError", "get")
def test_application_view_error(target_application):
    target_application.get("/view_error/", status=500)


_test_application_view_handle_error_scoped_metrics = list(_scoped_metrics)
_test_application_view_handle_error_scoped_metrics.append(("Function/urls:ViewHandleError.get", 1))


@pytest.mark.parametrize("status,should_record", [(418, True), (200, False)])
@pytest.mark.parametrize("use_global_exc_handler", [True, False])
@validate_code_level_metrics("urls.ViewHandleError", "get")
def test_application_view_handle_error(status, should_record, use_global_exc_handler, target_application):
    errors = ["urls:Error"] if should_record else []

    @validate_transaction_errors(errors=errors)
    @validate_transaction_metrics(
        "urls:ViewHandleError.get", scoped_metrics=_test_application_view_handle_error_scoped_metrics
    )
    def _test():
        response = target_application.get("/view_handle_error/%s/%s/" % (status, use_global_exc_handler), status=status)
        if use_global_exc_handler:
            response.mustcontain("exception was handled global")
        else:
            response.mustcontain("exception was handled not global")

    _test()


_test_api_view_view_name_get = "urls:wrapped_view.get"
_test_api_view_scoped_metrics_get = list(_scoped_metrics)
_test_api_view_scoped_metrics_get.append(("Function/%s" % _test_api_view_view_name_get, 1))


@validate_transaction_errors(errors=[])
@validate_transaction_metrics(_test_api_view_view_name_get, scoped_metrics=_test_api_view_scoped_metrics_get)
@validate_code_level_metrics("urls.WrappedAPIView", "wrapped_view", py2_namespace="urls")
def test_api_view_get(target_application):
    response = target_application.get("/api_view/")
    response.mustcontain("wrapped_view response")


_test_api_view_view_name_post = "urls:wrapped_view.http_method_not_allowed"
_test_api_view_scoped_metrics_post = list(_scoped_metrics)
_test_api_view_scoped_metrics_post.append(("Function/%s" % _test_api_view_view_name_post, 1))


@validate_transaction_errors(errors=["rest_framework.exceptions:MethodNotAllowed"])
@validate_transaction_metrics(_test_api_view_view_name_post, scoped_metrics=_test_api_view_scoped_metrics_post)
def test_api_view_method_not_allowed(target_application):
    target_application.post("/api_view/", status=405)


def test_application_view_agent_disabled(target_application):
    settings = global_settings()

    @override_generic_settings(settings, {"enabled": False})
    @function_not_called("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _test():
        response = target_application.get("/view/")
        assert response.status_int == 200
        response.mustcontain("restframework view response")

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio

import pytest
from testing_support.fixtures import (
    dt_enabled,
    override_application_settings,
    override_generic_settings,
)
from testing_support.mock_external_http_server import (
    MockExternalHTTPHResponseHeadersServer,
)
from testing_support.validators.validate_cross_process_headers import (
    validate_cross_process_headers,
)
from testing_support.validators.validate_span_events import validate_span_events
from testing_support.validators.validate_transaction_errors import (
    validate_transaction_errors,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_segment_params import (
    validate_tt_segment_params,
)

from newrelic.api.background_task import background_task
from newrelic.api.time_trace import current_trace
from newrelic.api.transaction import current_transaction
from newrelic.core.config import global_settings

ENCODING_KEY = "1234567890123456789012345678901234567890"
SCOPED_METRICS = []
ROLLUP_METRICS = [("External/all", 2), ("External/allOther", 2)]

CAT_RESPONSE_CODE = None


def cat_response_handler(self):
    global CAT_RESPONSE_CODE
    if not CAT_RESPONSE_CODE:
        raise ValueError("CAT_RESPONSE_CODE must be a valid status_code.")

    # payload
    # (
    #     u'1#1', u'WebTransaction/Function/app:beep',
    #     0, 1.23, -1,
    #     'dd4a810b7cb7f937',
    #     False,
    # )
    cat_response_header = (
        "X-NewRelic-App-Data",
        "ahACFwQUGxpuVVNmQVVbRVZbTVleXBxyQFhUTFBfXx1SREUMVV1cQBMeAxgEGAULFR0AHhFQUQJWAAgAUwVQVgJQDgsOEh1UUlhGU2o=",
    )
    self.send_response(CAT_RESPONSE_CODE)
    self.send_header(*cat_response_header)
    self.end_headers()
    self.wfile.write(b"Example Data")


@pytest.fixture(scope="session")
def server():
    external = MockExternalHTTPHResponseHeadersServer(handler=cat_response_handler)
    with external:
        yield external


@pytest.fixture()
def populate_metrics(server, request):
    SCOPED_METRICS[:] = []
    method = request.getfixturevalue("method").upper()
    SCOPED_METRICS.append(("External/localhost:%d/httpx/%s" % (server.port, method), 2))


def exercise_sync_client(server, client, method):
    with client as client:
        resolved_method = getattr(client, method)
        resolved_method("http://localhost:%s" % server.port)
        response = resolved_method("http://localhost:%s" % server.port)

    return response


@pytest.mark.parametrize(
    "method",
    (
        "get",
        "options",
        "head",
        "post",
        "put",
        "patch",
        "delete",
    ),
)
@validate_transaction_metrics(
    "test_sync_client",
    scoped_metrics=SCOPED_METRICS,
    rollup_metrics=ROLLUP_METRICS,
    background_task=True,
)
@background_task(name="test_sync_client")
def test_sync_client(httpx, server, method):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = 200

    assert exercise_sync_client(server, httpx.Client(), method).status_code == 200


async def exercise_async_client(server, client, method):
    async with client as client:
        resolved_method = getattr(client, method)
        responses = await asyncio.gather(
            resolved_method("http://localhost:%s" % server.port),
            resolved_method("http://localhost:%s" % server.port),
        )

    return responses


@pytest.mark.parametrize(
    "method",
    (
        "get",
        "options",
        "head",
        "post",
        "put",
        "patch",
        "delete",
    ),
)
@validate_transaction_metrics(
    "test_async_client",
    scoped_metrics=SCOPED_METRICS,
    rollup_metrics=ROLLUP_METRICS,
    background_task=True,
)
@background_task(name="test_async_client")
def test_async_client(httpx, server, loop, method):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = 200

    responses = loop.run_until_complete(exercise_async_client(server, httpx.AsyncClient(), method))
    assert all(response.status_code == 200 for response in responses)


@pytest.mark.parametrize(
    "distributed_tracing,span_events",
    (
        (True, True),
        (True, False),
        (False, False),
    ),
)
def test_sync_cross_process_request(httpx, server, distributed_tracing, span_events):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = 200

    @override_application_settings(
        {
            "distributed_tracing.enabled": distributed_tracing,
            "span_events.enabled": span_events,
            "cross_application_tracer.enabled": not distributed_tracing,
        }
    )
    @validate_transaction_errors(errors=[])
    @background_task(name="test_sync_cross_process_request")
    @validate_cross_process_headers
    def _test():
        transaction = current_transaction()

        with httpx.Client() as client:
            response = client.get("http://localhost:%s" % server.port)

        transaction._test_request_headers = response.request.headers

        assert response.status_code == 200

    _test()


@pytest.mark.parametrize(
    "distributed_tracing,span_events",
    (
        (True, True),
        (True, False),
        (False, False),
    ),
)
@validate_transaction_errors(errors=[])
@background_task(name="test_async_cross_process_request")
@validate_cross_process_headers
def test_async_cross_process_request(httpx, server, loop, distributed_tracing, span_events):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = 200

    @override_application_settings(
        {
            "distributed_tracing.enabled": distributed_tracing,
            "span_events.enabled": span_events,
        }
    )
    async def _test():
        async with httpx.AsyncClient() as client:
            response = await client.get("http://localhost:%s" % server.port)

        return response

    transaction = current_transaction()
    response = loop.run_until_complete(_test())
    transaction._test_request_headers = response.request.headers

    assert response.status_code == 200


@override_application_settings(
    {
        "distributed_tracing.enabled": True,
        "span_events.enabled": True,
        "cross_application_tracer.enabled": True,
    }
)
@validate_transaction_errors(errors=[])
@background_task(name="test_sync_cross_process_override_headers")
def test_sync_cross_process_override_headers(httpx, server, loop):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = 200

    transaction = current_transaction()

    with httpx.Client() as client:
        response = client.get("http://localhost:%s" % server.port, headers={"newrelic": "1234"})

    transaction._test_request_headers = response.request.headers

    assert response.status_code == 200
    assert response.request.headers["newrelic"] == "1234"


@override_application_settings(
    {
        "distributed_tracing.enabled": True,
        "span_events.enabled": True,
    }
)
@validate_transaction_errors(errors=[])
@background_task(name="test_async_cross_process_override_headers")
def test_async_cross_process_override_headers(httpx, server, loop):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = 200

    async def _test():
        async with httpx.AsyncClient() as client:
            response = await client.get("http://localhost:%s" % server.port, headers={"newrelic": "1234"})

        return response

    response = loop.run_until_complete(_test())

    assert response.status_code == 200
    assert response.request.headers["newrelic"] == "1234"


@pytest.mark.parametrize("cat_enabled", [True, False])
@pytest.mark.parametrize("response_code", [200, 500])
def test_sync_client_cat_response_processing(cat_enabled, response_code, server, httpx):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = response_code

    _custom_settings = {
        "cross_process_id": "1#1",
        "encoding_key": ENCODING_KEY,
        "trusted_account_ids": [1],
        "cross_application_tracer.enabled": cat_enabled,
        "distributed_tracing.enabled": False,
        "transaction_tracer.transaction_threshold": 0.0,
    }

    expected_metrics = [
        (
            "ExternalTransaction/localhost:%s/1#1/WebTransaction/Function/app:beep" % server.port,
            1 if cat_enabled else None,
        ),
    ]

    @validate_transaction_metrics(
        "test_sync_client_cat_response_processing",
        background_task=True,
        rollup_metrics=expected_metrics,
        scoped_metrics=expected_metrics,
    )
    @validate_tt_segment_params(exact_params={"http.statusCode": response_code})
    @override_application_settings(_custom_settings)
    @background_task(name="test_sync_client_cat_response_processing")
    def _test():
        with httpx.Client() as client:
            response = client.get("http://localhost:%s" % server.port)

    _test()


@pytest.mark.parametrize("cat_enabled", [True, False])
@pytest.mark.parametrize("response_code", [200, 500])
def test_async_client_cat_response_processing(cat_enabled, response_code, httpx, server, loop):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = response_code

    _custom_settings = {
        "cross_process_id": "1#1",
        "encoding_key": ENCODING_KEY,
        "trusted_account_ids": [1],
        "cross_application_tracer.enabled": cat_enabled,
        "distributed_tracing.enabled": False,
        "transaction_tracer.transaction_threshold": 0.0,
    }

    expected_metrics = [
        (
            "ExternalTransaction/localhost:%s/1#1/WebTransaction/Function/app:beep" % server.port,
            1 if cat_enabled else None,
        ),
    ]

    @validate_transaction_metrics(
        "test_async_client_cat_response_processing",
        background_task=True,
        rollup_metrics=expected_metrics,
        scoped_metrics=expected_metrics,
    )
    @validate_tt_segment_params(exact_params={"http.statusCode": response_code})
    @override_application_settings(_custom_settings)
    @background_task(name="test_async_client_cat_response_processing")
    def _test():
        async def coro():
            async with httpx.AsyncClient() as client:
                response = await client.get("http://localhost:%s" % server.port)

            return response

        response = loop.run_until_complete(coro())

    _test()


@dt_enabled
def test_sync_client_event_hook_exception(httpx, server):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = 500

    def exception_event_hook(response):
        if response.status_code >= 400:
            raise RuntimeError()

    def empty_hook(response):
        pass

    @validate_span_events(
        count=1,
        exact_intrinsics={"name": "External/localhost:%d/httpx/GET" % server.port},
        exact_agents={"http.statusCode": CAT_RESPONSE_CODE},
    )
    @background_task(name="test_sync_client_event_hook_exception")
    def make_request(client, exc_expected=True):
        if exc_expected:
            with pytest.raises(RuntimeError):
                client.get("http://localhost:%s" % server.port)
        else:
            client.get("http://localhost:%s" % server.port)

    with httpx.Client(event_hooks={"response": [exception_event_hook]}) as client:
        # Test client init
        make_request(client)

        # Test client setter
        client.event_hooks = {"response": [exception_event_hook]}
        make_request(client)

        # Test dict setitem
        client.event_hooks["response"] = [exception_event_hook]
        make_request(client)

        # Test list insert
        client.event_hooks["response"].insert(0, exception_event_hook)
        make_request(client)

        # Don't crash if response isn't specified
        client.event_hooks = {"request": [empty_hook]}
        make_request(client, exc_expected=False)


@override_application_settings({"distributed_tracing.enabled": True, "span_events.enabled": True})
def test_async_client_event_hook_exception(httpx, server, loop):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = 500

    def exception_event_hook(response):
        if response.status_code >= 400:
            raise RuntimeError()

    def empty_hook(response):
        pass

    @validate_span_events(
        count=1,
        exact_intrinsics={"name": "External/localhost:%d/httpx/GET" % server.port},
        exact_agents={"http.statusCode": CAT_RESPONSE_CODE},
    )
    @background_task(name="test_sync_client_event_hook_exception")
    def make_request(client, exc_expected=True):
        async def coro():
            if exc_expected:
                with pytest.raises(RuntimeError):
                    await client.get("http://localhost:%s" % server.port)
            else:
                await client.get("http://localhost:%s" % server.port)

        loop.run_until_complete(coro())

    def _test():
        with httpx.AsyncClient(event_hooks={"response": [exception_event_hook]}) as client:
            # Test client init
            make_request(client)

            # Test client setter
            client.event_hooks = {"response": [exception_event_hook]}
            make_request(client)

            # Test dict setitem
            client.event_hooks["response"] = [exception_event_hook]
            make_request(client)

            # Test list insert
            client.event_hooks["response"].insert(0, exception_event_hook)
            make_request(client)

            # Don't crash if response isn't specified
            client.event_hooks = {"request": [empty_hook]}
            make_request(client, exc_expected=False)


@override_generic_settings(
    global_settings(),
    {
        "enabled": False,
    },
)
def test_sync_nr_disabled(httpx, server):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = 200

    with httpx.Client() as client:
        trace = current_trace()
        response = client.get("http://localhost:%s" % server.port)

        assert response.status_code == 200
        assert trace is None


@override_generic_settings(
    global_settings(),
    {
        "enabled": False,
    },
)
def test_async_nr_disabled(httpx, server, loop):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = 200

    async def _test():
        async with httpx.AsyncClient() as client:
            response = await client.get("http://localhost:%s" % server.port)

        return response

    trace = current_trace()
    response = loop.run_until_complete(_test())
    assert response.status_code == 200
    assert trace is None


@pytest.mark.parametrize(
    "client",
    (
        "Client",
        "AsyncClient",
    ),
)
def test_invalid_import_order_client(monkeypatch, httpx, server, loop, client):
    global CAT_RESPONSE_CODE
    CAT_RESPONSE_CODE = 200

    if "Async" in client:
        is_async = True
    else:
        is_async = False

    client = getattr(httpx, client)

    # Force the client class into the state as if instrumentation had not run
    monkeypatch.setattr(client, "_event_hooks", None)

    # Instantiate a client
    client = client()

    # Remove monkeypatching of _event_hooks to restore instrumentation
    monkeypatch.undo()

    if is_async:
        responses = loop.run_until_complete(exercise_async_client(server, client, "get"))
        assert all(response.status_code == 200 for response in responses)
    else:
        response = exercise_sync_client(server, client, "get")
        assert response.status_code == 200



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aredis
import pytest
from testing_support.db_settings import redis_settings
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_metrics import validate_transaction_metrics
from testing_support.util import instance_hostname

from newrelic.api.background_task import background_task

DB_MULTIPLE_SETTINGS = redis_settings()

# Settings

_enable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": True,
}
_disable_instance_settings = {
    "datastore_tracer.instance_reporting.enabled": False,
}

# Metrics

_base_scoped_metrics = (
    ("Datastore/operation/Redis/get", 1),
    ("Datastore/operation/Redis/set", 1),
    ("Datastore/operation/Redis/client_list", 1),
)

_base_rollup_metrics = (
    ("Datastore/all", 3),
    ("Datastore/allOther", 3),
    ("Datastore/Redis/all", 3),
    ("Datastore/Redis/allOther", 3),
    ("Datastore/operation/Redis/get", 1),
    ("Datastore/operation/Redis/get", 1),
    ("Datastore/operation/Redis/client_list", 1),
)

_concurrent_scoped_metrics = [
    ("Datastore/operation/Redis/get", 2),
    ("Datastore/operation/Redis/set", 2),
]

_concurrent_rollup_metrics = [
    ("Datastore/all", 4),
    ("Datastore/allOther", 4),
    ("Datastore/Redis/all", 4),
    ("Datastore/Redis/allOther", 4),
    ("Datastore/operation/Redis/set", 2),
    ("Datastore/operation/Redis/get", 2),
]

_disable_scoped_metrics = list(_base_scoped_metrics)
_disable_rollup_metrics = list(_base_rollup_metrics)

_enable_scoped_metrics = list(_base_scoped_metrics)
_enable_rollup_metrics = list(_base_rollup_metrics)

if len(DB_MULTIPLE_SETTINGS) > 1:
    redis_1 = DB_MULTIPLE_SETTINGS[0]
    redis_2 = DB_MULTIPLE_SETTINGS[1]

    host_1 = instance_hostname(redis_1["host"])
    port_1 = redis_1["port"]

    host_2 = instance_hostname(redis_2["host"])
    port_2 = redis_2["port"]

    instance_metric_name_1 = "Datastore/instance/Redis/%s/%s" % (host_1, port_1)
    instance_metric_name_2 = "Datastore/instance/Redis/%s/%s" % (host_2, port_2)

    _enable_rollup_metrics.extend(
        [
            (instance_metric_name_1, 2),
            (instance_metric_name_2, 1),
        ]
    )

    _disable_rollup_metrics.extend(
        [
            (instance_metric_name_1, None),
            (instance_metric_name_2, None),
        ]
    )

    _concurrent_rollup_metrics.extend(
        [
            (instance_metric_name_1, 2),
            (instance_metric_name_2, 2),
        ]
    )


async def exercise_redis(client_1, client_2):
    await client_1.set("key", "value")
    await client_1.get("key")

    await client_2.execute_command("CLIENT", "LIST", parse="LIST")


@pytest.mark.skipif(len(DB_MULTIPLE_SETTINGS) < 2, reason="Test environment not configured with multiple databases.")
@override_application_settings(_enable_instance_settings)
@validate_transaction_metrics(
    "test_multiple_dbs:test_multiple_datastores_enabled",
    scoped_metrics=_enable_scoped_metrics,
    rollup_metrics=_enable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_multiple_datastores_enabled(loop):
    redis1 = DB_MULTIPLE_SETTINGS[0]
    redis2 = DB_MULTIPLE_SETTINGS[1]

    client_1 = aredis.StrictRedis(host=redis1["host"], port=redis1["port"], db=0)
    client_2 = aredis.StrictRedis(host=redis2["host"], port=redis2["port"], db=1)
    loop.run_until_complete(exercise_redis(client_1, client_2))


@pytest.mark.skipif(len(DB_MULTIPLE_SETTINGS) < 2, reason="Test environment not configured with multiple databases.")
@override_application_settings(_disable_instance_settings)
@validate_transaction_metrics(
    "test_multiple_dbs:test_multiple_datastores_disabled",
    scoped_metrics=_disable_scoped_metrics,
    rollup_metrics=_disable_rollup_metrics,
    background_task=True,
)
@background_task()
def test_multiple_datastores_disabled(loop):
    redis1 = DB_MULTIPLE_SETTINGS[0]
    redis2 = DB_MULTIPLE_SETTINGS[1]

    client_1 = aredis.StrictRedis(host=redis1["host"], port=redis1["port"], db=0)
    client_2 = aredis.StrictRedis(host=redis2["host"], port=redis2["port"], db=1)
    loop.run_until_complete(exercise_redis(client_1, client_2))


@pytest.mark.skipif(len(DB_MULTIPLE_SETTINGS) < 2, reason="Test environment not configured with multiple databases.")
@validate_transaction_metrics(
    "test_multiple_dbs:test_concurrent_calls",
    scoped_metrics=_concurrent_scoped_metrics,
    rollup_metrics=_concurrent_rollup_metrics,
    background_task=True,
)
@override_application_settings(_enable_instance_settings)
@background_task()
def test_concurrent_calls(loop):
    # Concurrent calls made with original instrumenation taken from synchonous Redis
    # instrumentation had a bug where datastore info on concurrent calls to multiple instances
    # would result in all instances reporting as the host/port of the final call made.

    import asyncio

    redis1 = DB_MULTIPLE_SETTINGS[0]
    redis2 = DB_MULTIPLE_SETTINGS[1]

    client_1 = aredis.StrictRedis(host=redis1["host"], port=redis1["port"], db=0)
    client_2 = aredis.StrictRedis(host=redis2["host"], port=redis2["port"], db=1)
    clients = (client_1, client_2)

    async def exercise_concurrent():
        await asyncio.gather(*(client.set("key-%d" % i, i) for i, client in enumerate(clients)))
        await asyncio.gather(*(client.get("key-%d" % i) for i, client in enumerate(clients)))

    loop.run_until_complete(exercise_concurrent())



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webtest

from flask import Flask

application = Flask(__name__)

@application.before_first_request
def before_first_request():
    pass

@application.before_request
def before_request():
    pass

@application.after_request
def after_request(response):
    return response

@application.teardown_request
def teardown_request(exc):
    pass

@application.teardown_appcontext
def teardown_appcontext(exc):
    pass

@application.route('/middleware')
def index_page():
    return 'INDEX RESPONSE'

_test_application = webtest.TestApp(application)



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from newrelic.common.encoding_utils import unpack_field
from newrelic.common.object_wrapper import transient_function_wrapper
from newrelic.common.system_info import LOCALHOST_EQUIVALENTS
from newrelic.core.database_utils import SQLConnections
from newrelic.packages import six


def validate_slow_sql_collector_json(required_params=set(),
        forgone_params=set(), exact_params=None):
    """Check that slow_sql json output is in accordance with agent specs.
    """
    @transient_function_wrapper('newrelic.core.stats_engine',
            'StatsEngine.record_transaction')
    def _validate_slow_sql_collector_json(wrapped, instance, args, kwargs):
        legal_param_keys = set([
            'explain_plan',
            'backtrace',
            'host',
            'port_path_or_id',
            'database_name',
            'parent.type',
            'parent.app',
            'parent.account',
            'parent.transportType',
            'parent.transportDuration',
            'guid',
            'traceId',
            'priority',
            'sampled',
        ])
        try:
            result = wrapped(*args, **kwargs)
        except:
            raise
        else:
            connections = SQLConnections()
            slow_sql_list = instance.slow_sql_data(connections)

            for slow_sql in slow_sql_list:
                assert isinstance(slow_sql[0], six.string_types)  # txn_name
                assert isinstance(slow_sql[1], six.string_types)  # txn_url
                assert isinstance(slow_sql[2], int)               # sql_id
                assert isinstance(slow_sql[3], six.string_types)  # sql
                assert isinstance(slow_sql[4], six.string_types)  # metric_name
                assert isinstance(slow_sql[5], int)               # count
                assert isinstance(slow_sql[6], float)             # total
                assert isinstance(slow_sql[7], float)             # min
                assert isinstance(slow_sql[8], float)             # max
                assert isinstance(slow_sql[9], six.string_types)  # params

                params = slow_sql[9]
                data = unpack_field(params)

                # only legal keys should be reported
                assert len(set(data.keys()) - legal_param_keys) == 0

                # if host is reported, it cannot be localhost
                if 'host' in data:
                    assert data['host'] not in LOCALHOST_EQUIVALENTS

                if required_params:
                    for param in required_params:
                        assert param in data

                if forgone_params:
                    for param in forgone_params:
                        assert param not in data

                if exact_params:
                    for param, value in exact_params.items():
                        assert param in data
                        assert data[param] == value

        return result

    return _validate_slow_sql_collector_json



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import psycopg2
try:
    from psycopg2 import sql
except ImportError:
    sql = None

from testing_support.db_settings import postgresql_settings
from newrelic.api.background_task import background_task

DB_SETTINGS = postgresql_settings()[0]


@pytest.fixture(scope='module')
def conn():
    conn = psycopg2.connect(
            database=DB_SETTINGS['name'], user=DB_SETTINGS['user'],
            password=DB_SETTINGS['password'], host=DB_SETTINGS['host'],
            port=DB_SETTINGS['port'])
    yield conn
    conn.close()


@background_task()
def test_as_string_1(conn):

    # All of these are similar to those described in the doctests in
    # psycopg2/lib/sql.py

    comp = sql.Composed(
            [sql.SQL("insert into "), sql.Identifier("table")])
    result = comp.as_string(conn)
    assert result == 'insert into "table"'


@background_task()
def test_as_string_2(conn):
    fields = sql.Identifier('foo') + sql.Identifier('bar')  # a Composed
    result = fields.join(', ').as_string(conn)
    assert result == '"foo", "bar"'


@background_task()
def test_as_string_3(conn):
    query = sql.SQL("select {0} from {1}").format(
        sql.SQL(', ').join([sql.Identifier('foo'), sql.Identifier('bar')]),
        sql.Identifier('table'))
    result = query.as_string(conn)
    assert result == 'select "foo", "bar" from "table"'


@background_task()
def test_as_string_4(conn):
    result = sql.SQL("select * from {0} where {1} = %s").format(
            sql.Identifier('people'), sql.Identifier('id')).as_string(conn)
    assert result == 'select * from "people" where "id" = %s'


@background_task()
def test_as_string_5(conn):
    result = sql.SQL("select * from {tbl} where {pkey} = %s").format(
            tbl=sql.Identifier('people'), pkey=sql.Identifier('id')).as_string(
            conn)
    assert result == 'select * from "people" where "id" = %s'


@background_task()
def test_as_string_6(conn):
    snip = sql.SQL(', ').join(
            sql.Identifier(n) for n in ['foo', 'bar', 'baz'])
    result = snip.as_string(conn)
    assert result == '"foo", "bar", "baz"'


@background_task()
def test_as_string_7(conn):
    t1 = sql.Identifier("foo")
    t2 = sql.Identifier("ba'r")
    t3 = sql.Identifier('ba"z')
    result = sql.SQL(', ').join([t1, t2, t3]).as_string(conn)
    assert result == '"foo", "ba\'r", "ba""z"'


@background_task()
def test_as_string_8(conn):
    s1 = sql.Literal("foo")
    s2 = sql.Literal("ba'r")
    s3 = sql.Literal(42)
    result = sql.SQL(', ').join([s1, s2, s3]).as_string(conn)
    assert result == "'foo', 'ba''r', 42"


@background_task()
def test_as_string_9(conn):
    names = ['foo', 'bar', 'baz']
    q1 = sql.SQL("insert into table ({0}) values ({1})").format(
            sql.SQL(', ').join(map(sql.Identifier, names)),
            sql.SQL(', ').join(sql.Placeholder() * len(names)))
    result = q1.as_string(conn)
    assert (result ==
            'insert into table ("foo", "bar", "baz") values (%s, %s, %s)')


@background_task()
def test_as_string_10(conn):
    names = ['foo', 'bar', 'baz']
    q2 = sql.SQL("insert into table ({0}) values ({1})").format(
            sql.SQL(', ').join(map(sql.Identifier, names)),
            sql.SQL(', ').join(map(sql.Placeholder, names)))
    result = q2.as_string(conn)
    assert (result ==
        'insert into table ("foo", "bar", "baz") '
        'values (%(foo)s, %(bar)s, %(baz)s)')



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import os
import random
import socket
import time
from testing_support.fixtures import TerminatingPopen
from testing_support.util import get_open_port


pytest.importorskip('asyncio')
from urllib.request import urlopen
from testing_support.util import get_open_port

@pytest.mark.parametrize('nr_enabled', (True, False))
def test_asgi_app(nr_enabled):
    nr_admin = os.path.join(os.environ['TOX_ENV_DIR'], 'bin', 'newrelic-admin')
    gunicorn = os.path.join(os.environ['TOX_ENV_DIR'], 'bin', 'gunicorn')

    PORT = get_open_port()
    cmd = [gunicorn, '-b', '127.0.0.1:%d' % PORT, '--worker-class',
            'worker.AsgiWorker', 'asgi_app:Application']

    if nr_enabled:
        env = {
            'NEW_RELIC_ENABLED': 'true',
            'NEW_RELIC_HOST': 'staging-collector.newrelic.com',
            'NEW_RELIC_APP_NAME': 'Python Agent Test (gunicorn)',
            'NEW_RELIC_CONFIG_FILE': 'config.ini',
            'NEW_RELIC_LOG': 'stderr',
            'NEW_RELIC_LOG_LEVEL': 'debug',
            'NEW_RELIC_STARTUP_TIMEOUT': '10.0',
            'NEW_RELIC_SHUTDOWN_TIMEOUT': '10.0',
        }
        new_cmd = [nr_admin, 'run-program']
        new_cmd.extend(cmd)
        cmd = new_cmd
    else:
        env = {}

    # Restart the server if it dies during testing
    for _ in range(5):
        with TerminatingPopen(cmd, env=env):
            for _ in range(50):
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                try:
                    s.connect(('127.0.0.1', PORT))
                    s.close()
                    break
                except socket.error:
                    pass

                time.sleep(0.1)
            else:
                continue
            with urlopen('http://127.0.0.1:%d' % PORT) as resp:
                assert resp.getcode() == 200
                assert resp.read() == b'PONG'

        # test passed
        break
    else:
        assert False, 'Gunicorn test did not run'



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_tt_segment_params import (
    validate_tt_segment_params,
)

from newrelic.api.background_task import background_task
from newrelic.api.database_trace import DatabaseTrace
from newrelic.api.datastore_trace import DatastoreTrace
from newrelic.api.external_trace import ExternalTrace, external_trace
from newrelic.api.function_trace import FunctionTrace
from newrelic.api.graphql_trace import GraphQLOperationTrace, GraphQLResolverTrace
from newrelic.api.memcache_trace import MemcacheTrace
from newrelic.api.message_trace import MessageTrace
from newrelic.api.solr_trace import SolrTrace
from newrelic.api.transaction import current_transaction


@external_trace("lib", "https://example.com/path?q=q#frag")
def external():
    pass


@validate_tt_segment_params(present_params=("http.url",))
@background_task(name="test_external_segment_attributes_default")
def test_external_segment_attributes_default():
    external()


@override_application_settings(
    {
        "transaction_segments.attributes.exclude": ["http.url"],
    }
)
@validate_tt_segment_params(forgone_params=("http.url",))
@background_task(name="test_external_segment_attributes_disabled")
def test_external_segment_attributes_disabled():
    external()


@validate_tt_segment_params(exact_params={"http.url": "http://example.org"})
@background_task(name="test_external_user_params_override_url")
def test_external_user_params_override_url():
    with ExternalTrace("lib", "http://example.com") as t:
        # Pretend like this is a user attribute and it's legal to do this
        t.params["http.url"] = "http://example.org"


@validate_tt_segment_params(exact_params={"db.instance": "a" * 255})
@background_task(name="test_datastore_db_instance_truncation")
def test_datastore_db_instance_truncation():
    with DatastoreTrace("db_product", "db_target", "db_operation", database_name="a" * 256):
        pass


@validate_tt_segment_params(exact_params={"db.instance": "a" * 255})
@background_task(name="test_database_db_instance_truncation")
def test_database_db_instance_truncation():
    with DatabaseTrace("select * from foo", database_name="a" * 256):
        pass


@override_application_settings(
    {
        "transaction_tracer.record_sql": "raw",
    }
)
@validate_tt_segment_params(exact_params={"db.statement": "select 1"})
@background_task(name="test_database_db_statement")
def test_database_db_statement_default_enabled():
    with DatabaseTrace("select 1"):
        pass


@override_application_settings(
    {
        "transaction_tracer.record_sql": "raw",
        "agent_limits.sql_query_length_maximum": 1,
    }
)
@validate_tt_segment_params(exact_params={"db.statement": "a"})
@background_task(name="test_database_db_statement_truncation")
def test_database_db_statement_truncation():
    with DatabaseTrace("a" * 2):
        pass


@override_application_settings(
    {
        "transaction_segments.attributes.exclude": ["db.*"],
    }
)
@validate_tt_segment_params(forgone_params=("db.instance", "db.statement"))
@background_task(name="test_database_segment_attributes_disabled")
def test_database_segment_attributes_disabled():
    transaction = current_transaction()
    with DatabaseTrace("select 1", database_name="foo"):
        pass


@pytest.mark.parametrize(
    "trace_type,args",
    (
        (DatabaseTrace, ("select * from foo",)),
        (DatastoreTrace, ("db_product", "db_target", "db_operation")),
        (ExternalTrace, ("lib", "url")),
        (FunctionTrace, ("name",)),
        (GraphQLOperationTrace, ()),
        (GraphQLResolverTrace, ()),
        (MemcacheTrace, ("command",)),
        (MessageTrace, ("lib", "operation", "dst_type", "dst_name")),
        (SolrTrace, ("lib", "command")),
    ),
)
def test_each_segment_type(trace_type, args):
    @validate_tt_segment_params(exact_params={"blah": "bloo"})
    @override_application_settings(
        {
            "distributed_tracing.enabled": True,
            "span_events.enabled": True,
            "attributes.include": ["blah"],
        }
    )
    @background_task(name="test_each_segment_type")
    def _test():

        transaction = current_transaction()
        transaction._sampled = True

        with trace_type(*args) as trace:
            trace._add_agent_attribute("blah", "bloo")

    _test()


@override_application_settings(
    {
        "distributed_tracing.enabled": True,
        "span_events.enabled": True,
        "attributes.include": ["*"],
    }
)
@background_task(name="test_attribute_overrides")
def test_attribute_overrides():
    with FunctionTrace("test_attribute_overrides_trace") as trace:
        trace.exclusive = 0.1
        trace._add_agent_attribute("exclusive_duration_millis", 0.2)
        trace._add_agent_attribute("test_attr", "a")
        trace.add_custom_attribute("exclusive_duration_millis", 0.3)
        trace.add_custom_attribute("test_attr", "b")
        node = trace.create_node()

    params = node.get_trace_segment_params(current_transaction().settings)

    assert params["exclusive_duration_millis"] == 100
    assert params["test_attr"] == "b"



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

import pytest
from testing_support.fixtures import override_application_settings
from testing_support.validators.validate_transaction_event_attributes import (
    validate_transaction_event_attributes,
)

from newrelic.api.lambda_handler import lambda_handler

CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
FIXTURE_DIR = os.path.normpath(os.path.join(CURRENT_DIR, "fixtures"))
FIXTURE = os.path.join(FIXTURE_DIR, "lambda_event_source.json")
tests = {}
events = {}


def _load_tests():
    with open(FIXTURE, "r") as fh:
        for test in json.loads(fh.read()):
            test_name = test.pop("name")

            test_file = test_name + ".json"
            path = os.path.join(FIXTURE_DIR, "lambda_event_source", test_file)
            with open(path, "r") as fh:
                events[test_name] = json.loads(fh.read())

            tests[test_name] = test
    return tests.keys()


class Context(object):
    aws_request_id = "cookies"
    invoked_function_arn = "arn"
    function_name = "cats"
    function_version = "$LATEST"
    memory_limit_in_mb = 128


@lambda_handler()
def handler(event, context):
    return {
        "statusCode": "200",
        "body": "{}",
        "headers": {
            "Content-Type": "application/json",
            "Content-Length": 2,
        },
    }


# The lambda_hander has been deprecated for 3+ years
@pytest.mark.skip(reason="The lambda_handler has been deprecated")
@pytest.mark.parametrize("test_name", _load_tests())
def test_lambda_event_source(test_name):
    _exact = {"user": {}, "intrinsic": {}, "agent": {}}

    expected_arn = tests[test_name].get("aws.lambda.eventSource.arn", None)
    if expected_arn:
        _exact["agent"]["aws.lambda.eventSource.arn"] = expected_arn
    else:
        pytest.skip("Nothing to test!")
        return

    @override_application_settings({"attributes.include": ["aws.*"]})
    @validate_transaction_event_attributes({}, exact_attrs=_exact)
    def _test():
        handler(events[test_name], Context)

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from testing_support.fixtures import check_attributes

from newrelic.common.encoding_utils import unpack_field
from newrelic.common.object_wrapper import function_wrapper, transient_function_wrapper
from newrelic.core.database_utils import SQLConnections


def validate_transaction_trace_attributes(
    required_params=None, forgone_params=None, should_exist=True, url=None, index=-1
):
    required_params = required_params or {}
    forgone_params = forgone_params or {}

    trace_data = []

    @transient_function_wrapper("newrelic.core.stats_engine", "StatsEngine.record_transaction")
    def _validate_transaction_trace_attributes(wrapped, instance, args, kwargs):

        result = wrapped(*args, **kwargs)

        # Now that transaction has been recorded, generate
        # a transaction trace

        connections = SQLConnections()
        _trace_data = instance.transaction_trace_data(connections)
        trace_data.append(_trace_data)

        return result

    @function_wrapper
    def wrapper(wrapped, instance, args, kwargs):
        _new_wrapper = _validate_transaction_trace_attributes(wrapped)
        result = _new_wrapper(*args, **kwargs)

        _trace_data = trace_data[index]
        trace_data[:] = []

        if url is not None:
            trace_url = _trace_data[0][3]
            assert url == trace_url

        pack_data = unpack_field(_trace_data[0][4])
        assert len(pack_data) == 2
        assert len(pack_data[0]) == 5
        parameters = pack_data[0][4]

        assert "intrinsics" in parameters
        assert "userAttributes" in parameters
        assert "agentAttributes" in parameters

        check_attributes(parameters, required_params, forgone_params)

        return result

    return wrapper



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
import sys

from newrelic.api.error_trace import ErrorTrace
from newrelic.api.function_trace import FunctionTrace


def nr_coro_execute_name_wrapper(wrapped, result, set_name):
    @functools.wraps(wrapped)
    async def _nr_coro_execute_name_wrapper():
        result_ = await result
        set_name()
        return result_

    return _nr_coro_execute_name_wrapper()


def nr_coro_resolver_error_wrapper(wrapped, name, trace, ignore, result, transaction):
    @functools.wraps(wrapped)
    async def _nr_coro_resolver_error_wrapper():
        with trace:
            with ErrorTrace(ignore=ignore):
                try:
                    return await result
                except Exception:
                    transaction.set_transaction_name(name, "GraphQL", priority=15)
                    raise

    return _nr_coro_resolver_error_wrapper()


def nr_coro_resolver_wrapper(wrapped, trace, ignore, result):
    @functools.wraps(wrapped)
    async def _nr_coro_resolver_wrapper():
        with trace:
            with ErrorTrace(ignore=ignore):
                return await result

    return _nr_coro_resolver_wrapper()

def nr_coro_graphql_impl_wrapper(wrapped, trace, ignore, result):
    @functools.wraps(wrapped)
    async def _nr_coro_graphql_impl_wrapper():
        try:
            with ErrorTrace(ignore=ignore):
                result_ = await result
        except:
            trace.__exit__(*sys.exc_info())
            raise
        else:
            trace.__exit__(None, None, None)
            return result_


    return _nr_coro_graphql_impl_wrapper()


# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from testing_support.validators.validate_database_duration import (
    validate_database_duration,
)
from testing_support.validators.validate_transaction_metrics import (
    validate_transaction_metrics,
)
from testing_support.validators.validate_tt_collector_json import (
    validate_tt_collector_json,
)

from newrelic.api.background_task import background_task


@pytest.fixture(autouse=True)
def sample_data(collection):
    for x in range(1, 6):
        collection.add({"x": x})

    subcollection_doc = collection.document("subcollection")
    subcollection_doc.set({})
    subcollection_doc.collection("subcollection1").add({})


# ===== Query =====


@pytest.fixture()
def exercise_query(collection):
    def _exercise_query():
        query = collection.select("x").limit(10).order_by("x").where(field_path="x", op_string="<=", value=3)
        assert len(query.get()) == 3
        assert len([_ for _ in query.stream()]) == 3

    return _exercise_query


def test_firestore_query(exercise_query, collection, instance_info):
    _test_scoped_metrics = [
        ("Datastore/statement/Firestore/%s/stream" % collection.id, 1),
        ("Datastore/statement/Firestore/%s/get" % collection.id, 1),
    ]

    _test_rollup_metrics = [
        ("Datastore/operation/Firestore/get", 1),
        ("Datastore/operation/Firestore/stream", 1),
        ("Datastore/all", 2),
        ("Datastore/allOther", 2),
        ("Datastore/instance/Firestore/%s/%s" % (instance_info["host"], instance_info["port_path_or_id"]), 2),
    ]

    @validate_database_duration()
    @validate_transaction_metrics(
        "test_firestore_query",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_firestore_query")
    def _test():
        exercise_query()

    _test()


@background_task()
def test_firestore_query_generators(collection, assert_trace_for_generator):
    query = collection.select("x").where(field_path="x", op_string="<=", value=3)
    assert_trace_for_generator(query.stream)


def test_firestore_query_trace_node_datastore_params(exercise_query, instance_info):
    @validate_tt_collector_json(datastore_params=instance_info)
    @background_task()
    def _test():
        exercise_query()

    _test()


# ===== AggregationQuery =====


@pytest.fixture()
def exercise_aggregation_query(collection):
    def _exercise_aggregation_query():
        aggregation_query = collection.select("x").where(field_path="x", op_string="<=", value=3).count()
        assert aggregation_query.get()[0][0].value == 3
        assert [_ for _ in aggregation_query.stream()][0][0].value == 3

    return _exercise_aggregation_query


def test_firestore_aggregation_query(exercise_aggregation_query, collection, instance_info):
    _test_scoped_metrics = [
        ("Datastore/statement/Firestore/%s/stream" % collection.id, 1),
        ("Datastore/statement/Firestore/%s/get" % collection.id, 1),
    ]

    _test_rollup_metrics = [
        ("Datastore/operation/Firestore/get", 1),
        ("Datastore/operation/Firestore/stream", 1),
        ("Datastore/all", 2),
        ("Datastore/allOther", 2),
        ("Datastore/instance/Firestore/%s/%s" % (instance_info["host"], instance_info["port_path_or_id"]), 2),
    ]

    @validate_database_duration()
    @validate_transaction_metrics(
        "test_firestore_aggregation_query",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_firestore_aggregation_query")
    def _test():
        exercise_aggregation_query()

    _test()


@background_task()
def test_firestore_aggregation_query_generators(collection, assert_trace_for_generator):
    aggregation_query = collection.select("x").where(field_path="x", op_string="<=", value=3).count()
    assert_trace_for_generator(aggregation_query.stream)


def test_firestore_aggregation_query_trace_node_datastore_params(exercise_aggregation_query, instance_info):
    @validate_tt_collector_json(datastore_params=instance_info)
    @background_task()
    def _test():
        exercise_aggregation_query()

    _test()


# ===== CollectionGroup =====


@pytest.fixture()
def patch_partition_queries(monkeypatch, client, collection, sample_data):
    """
    Partitioning is not implemented in the Firestore emulator.

    Ordinarily this method would return a generator of Cursor objects. Each Cursor must point at a valid document path.
    To test this, we can patch the RPC to return 1 Cursor which is pointed at any document available.
    The get_partitions will take that and make 2 QueryPartition objects out of it, which should be enough to ensure
    we can exercise the generator's tracing.
    """
    from google.cloud.firestore_v1.types.document import Value
    from google.cloud.firestore_v1.types.query import Cursor

    subcollection = collection.document("subcollection").collection("subcollection1")
    documents = [d for d in subcollection.list_documents()]

    def mock_partition_query(*args, **kwargs):
        yield Cursor(before=False, values=[Value(reference_value=documents[0].path)])

    monkeypatch.setattr(client._firestore_api, "partition_query", mock_partition_query)
    yield


@pytest.fixture()
def exercise_collection_group(client, collection, patch_partition_queries):
    def _exercise_collection_group():
        collection_group = client.collection_group(collection.id)
        assert len(collection_group.get())
        assert len([d for d in collection_group.stream()])

        partitions = [p for p in collection_group.get_partitions(1)]
        assert len(partitions) == 2
        documents = []
        while partitions:
            documents.extend(partitions.pop().query().get())
        assert len(documents) == 6

    return _exercise_collection_group


def test_firestore_collection_group(exercise_collection_group, client, collection, instance_info):
    _test_scoped_metrics = [
        ("Datastore/statement/Firestore/%s/get" % collection.id, 3),
        ("Datastore/statement/Firestore/%s/stream" % collection.id, 1),
        ("Datastore/statement/Firestore/%s/get_partitions" % collection.id, 1),
    ]

    _test_rollup_metrics = [
        ("Datastore/operation/Firestore/get", 3),
        ("Datastore/operation/Firestore/stream", 1),
        ("Datastore/operation/Firestore/get_partitions", 1),
        ("Datastore/all", 5),
        ("Datastore/allOther", 5),
        ("Datastore/instance/Firestore/%s/%s" % (instance_info["host"], instance_info["port_path_or_id"]), 5),
    ]

    @validate_database_duration()
    @validate_transaction_metrics(
        "test_firestore_collection_group",
        scoped_metrics=_test_scoped_metrics,
        rollup_metrics=_test_rollup_metrics,
        background_task=True,
    )
    @background_task(name="test_firestore_collection_group")
    def _test():
        exercise_collection_group()

    _test()


@background_task()
def test_firestore_collection_group_generators(client, collection, assert_trace_for_generator, patch_partition_queries):
    collection_group = client.collection_group(collection.id)
    assert_trace_for_generator(collection_group.get_partitions, 1)


def test_firestore_collection_group_trace_node_datastore_params(exercise_collection_group, instance_info):
    @validate_tt_collector_json(datastore_params=instance_info)
    @background_task()
    def _test():
        exercise_collection_group()

    _test()



# Copyright 2010 New Relic, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from newrelic.api.datastore_trace import DatastoreTrace, wrap_datastore_trace
from newrelic.api.transaction import current_transaction
from newrelic.common.object_wrapper import (
    FunctionWrapper,
    wrap_function_wrapper,
    wrap_object,
)


def _instance_info(memcache_host):
    try:
        host = memcache_host.ip
        port_path_or_id = str(memcache_host.port)
    except AttributeError:
        host = "localhost"
        port_path_or_id = str(memcache_host.address)

    return (host, port_path_or_id, None)


def _nr_get_server_wrapper(wrapped, instance, args, kwargs):
    transaction = current_transaction()

    if transaction is None:
        return wrapped(*args, **kwargs)

    result = wrapped(*args, **kwargs)

    instance_info = (None, None, None)

    try:
        tracer_settings = transaction.settings.datastore_tracer
        host = result[0]

        # memcached does not set the db name attribute so there's
        # no need to check the db name reporting setting
        if tracer_settings.instance_reporting.enabled and host is not None:
            instance_info = _instance_info(host)
    except:
        instance_info = ("unknown", "unknown", None)

    transaction._nr_datastore_instance_info = instance_info

    return result


def MemcacheSingleWrapper(wrapped, product, target, operation, module):
    def _nr_datastore_trace_wrapper_(wrapped, instance, args, kwargs):
        transaction = current_transaction()

        if transaction is None:
            return wrapped(*args, **kwargs)

        transaction._nr_datastore_instance_info = (None, None, None)

        dt = DatastoreTrace(product, target, operation)

        with dt:
            result = wrapped(*args, **kwargs)

            instance_info = transaction._nr_datastore_instance_info
            (host, port_path_or_id, db) = instance_info
            dt.host = host
            dt.port_path_or_id = port_path_or_id

            return result

    return FunctionWrapper(wrapped, _nr_datastore_trace_wrapper_)


def wrap_memcache_single(module, object_path, product, target, operation):
    wrap_object(module, "Client.%s" % object_path, MemcacheSingleWrapper, (product, target, operation, module))


_memcache_client_methods = (
    "delete",
    "incr",
    "decr",
    "add",
    "append",
    "prepend",
    "replace",
    "set",
    "cas",
    "get",
    "gets",
)

_memcache_multi_methods = ("delete_multi", "get_multi", "set_multi", "get_stats", "get_slabs", "flush_all")


def instrument_memcache(module):
    wrap_function_wrapper(module, "Client._get_server", _nr_get_server_wrapper)

    for name in _memcache_client_methods:
        if hasattr(module.Client, name):
            wrap_memcache_single(module, name, product="Memcached", target=None, operation=name)

    for name in _memcache_multi_methods:
        if hasattr(module.Client, name):
            wrap_datastore_trace(module, "Client.%s" % name, product="Memcached", target=None, operation=name)
